# Copyright 2018 SuperDARN
# Author: Marina Schmidt
"""
This file contains classes for reading and writing of SuperDARN file types.
The file types that are supported:
    - Iqdat
    - Rawacf
    - Fitacf
    - Grid
    - Map

Classes:
--------
SDarnUtilities: utilites class that contains static methods for
SuperDARN file type checking
SDarnRead : Reads SuperDARN files types
SDarnWrite : writes Dmap Record structure into a SuperDARN file type

Exceptions:
-----------
SuperDARNFieldMissingError
SuperDARNExtraFieldError
SuperDARNDataTypeError

Future work
-----------
Persistent real-time methods
    methods that can be called repetitively for reading and writing
    to a data stream that would be used in real-time feeds.
Organization
    rethink public and private methods? <--- discussion
Parallelization

Notes
-----
DmapRead and DmapWrite are inherited by SDarnRead and SDarnWrite
"""
import logging

from typing import Union, List

from pydarn import (DmapRead, DmapWrite, superdarn_exceptions,
                    superdarn_formats, dmap2dict)

pydarn_log = logging.getLogger('pydarn')


class SDarnUtilities():
    """
    Utility class that contains static methods that does dictionary set
    calculations used for determining if there is any missing or extra
    fields in superDARN file types. Also, does data format type checks
    for SuperDARN file types.

    Static Methods
    ----------------
        dict_key_diff(dict1, dict2)
            Returns a set of the difference between dict1 and dict2 keys
        missing_field_check(file_struct_list, record, rec_num)
            Checks if there is any missing fields in the record from
            a list of possible file fields
        extra_field_check(file_struct_list, record, rec_num)
            Checks if there is any extra fields in the record from
            a list of possible file fields
        incorrect_types_check(file_struct_list, record)
            Checks if there is any incorrect types in the record from
            a list of possible file
            fields and their data type formats
        dict_list2set(dict_list)
            Converts a list of dictionaries to a set containing their keys

    Future Work:
        Add utilities for converting DMAP records to dictionaries containing
        only the values and dictionaries to DMAP records.

    """
    @staticmethod
    def dict_key_diff(dict1: Union[dict, set],
                      dict2: Union[dict, set]) -> set:
        """
        Determines the difference in the key set from the
        first dictionary to the second dictionary.
        ex) Let A = {a, b, c} and B = {d, a, b}
        Then A - B = {c}

        Parameters:
        -----------
        dict1 : dict or set
            dictionary or set to subtract from
        dict2 : dict or set
            dictionary or set subtracting from dict1

        Return:
        ------
        dict_diff : set
            difference between dict1 and dict2 keys or the sets
        """
        diff_dict = set(dict1) - set(dict2)
        return diff_dict

    # TODO: might be moved to utility class as a static method
    # Also, might not be needed if we do not use the data types
    # in the structure  dictionaries
    @staticmethod
    def dict_list2set(dict_list: List[dict]) -> set:
        """
        Converts a list of dictionaries to list of sets

        Parameters:
        -----------
        dict_list : list
            list of dictionaries

        Return:
        ------
        complete_set : set
            set containing all dictionary key from the list of dicts
        """
        # convert dictionaries to set to do some set magic
        sets = [set(dic) for dic in dict_list]
        # TODO: if data types don't matter in the structure format then
        # they can become sets instead of dictionaries.
        # create a complete set list
        # * - expands the list out into multiple set arguments
        # then the union operator creates it into a full unique set
        # example: s = [{'a','v'}, {'v','x'}] => set.union(*s)
        # = {'a', 'v', 'x'}
        complete_set = set.union(*sets)
        return complete_set

    @staticmethod
    def missing_field_check(file_struct_list: List[dict],
                            record: dict, rec_num: int):
        """
        Checks if any fields are missing from the record compared to the file
        structure.

        Parameters:
        -----------
        file_struct_list : List[dict]
            List of dictionaries for the possible file structure fields
        record : dict
            Dictionary representing the dmap record
        rec_num : int
            Record number for better error message information

        Raises:
        -------
        SuperDARNFieldMissing
        """

        missing_fields = set()
        """
        We have to check each possible subset to see if there is a missing
        field in the subset or if the whole subset is missing. If there
        is a missing field then raise a SuperDARNFieldMissingError.
        If a full subset is missing then this is technically okay because
        RST will add fields, or do not included fields based on the commands,
        and options the user uses to process the data.
        Some fields are not included if the data is not good quality,
        this occurs in the fitting FITACF files types due to the fitting
        procedure in fitacf 2.5 and 3.0.
        """

        for file_struct in file_struct_list:
            diff_fields = \
                SDarnUtilities.dict_key_diff(file_struct,
                                             record)
            # If 0 nothing missing, if len(file_struct) then
            # that subset is missing only meaning that command option was
            # not used, not necessarily meaning that it is a record.
            if len(diff_fields) not in (0, len(file_struct)):
                missing_fields = missing_fields.union(diff_fields)

        if len(missing_fields) > 0:
            raise superdarn_exceptions.\
                SuperDARNFieldMissingError(rec_num, missing_fields)

    @staticmethod
    def extra_field_check(file_struct_list: List[dict],
                          record: dict, rec_num: int):
        """
        Check if there is an extra field in the file structure list and record.

        Parameters
        -----------
        file_struct_list : List[dict]
            List of dictionary representing the possible fields
            in file structure
        record : dict
            Dmap record
        rec_num : int
            Record number for better error message information

        Raises
        -------
        SuperDARNFieldExtra

        """
        file_struct = SDarnUtilities.dict_list2set(file_struct_list)
        extra_fields = SDarnUtilities.dict_key_diff(record, file_struct)

        if len(extra_fields) > 0:
            raise superdarn_exceptions.SuperDARNExtraFieldError(rec_num,
                                                                extra_fields)

    # TODO: Do we want to keep this? If not, then change
    # SuperDARN_format_structure types to sets and get rid
    # of dict_list2set method.
    @staticmethod
    def incorrect_types_check(file_struct_list: List[dict], record: dict,
                              rec_num: int):
        """
        Checks if the file structure fields data type formats are correct
        in the record.

        Parameters
        ----------
        file_struct_list : List[dict]
            List of dictionaries representing the possible fields
            in a file structure
        record : dict
            Dmap record
        rec_num : int
            Record number for a better error message information

        Raises
        ------
        SuperDARNFileFormatError
        """
        complete_dict = {}
        for file_struct in file_struct_list:
            complete_dict.update(file_struct)
        incorrect_types_check = {param: complete_dict[param]
                                 for param in record.keys()
                                 if record[param].data_type_fmt != complete_dict[param]}
        if len(incorrect_types_check) > 0:
            raise superdarn_exceptions.\
                SuperDARNDataFormatTypeError(incorrect_types_check, rec_num)


class SDarnRead(DmapRead):
    """
    Reading and testing the integrity of SuperDARN file/stream types.
    Currently support file types are:
        Iqdat
        Rawacf
        Fitacf
        Grid
        Map
    ...

    Attributes
    ----------
    filename : str
        DMAP file name or data stream (give data_stream=True)
    cursor : int
        Current position in the byte array
    dmap_end_bytes : int
        The length of the byte array
    rec_num : int
        The record number in the DMAP format, helps with better error messages
    dmap_records: list[namedtuple]
        Dmap structure of records containing all dmap information
    records: list[dict]
        records containing a dictionary of the fields

    Methods
    -------
    read_iqdat()
        reads and checks iqdat DMAP binary data
    read_rawacf()
        reads and checks rawacf DMAP binary data
    read_fitacf()
        reads and checks fitacf DMAP binary data
    read_grid()
        reads and checks grid DMAP binary data
    read_map()
        reads and checks map DMAP binary data

    See Also
    --------
    DmapRead
        Class it inherits from
    read_records()
        reads DMAP binary data
    """
    def __init__(self, filename: str, stream: bool = False):
        """
        Extension of DmapRead constructor that reads SuperDARN file/stream type
        into a byte array for reading methods.

        Parameters
        ----------
        filename : str or bytes
            file name or data stream (given data_stream=True)
            containing dmap data.
        stream : bool
            default to false, boolean that indicates if dmap_file is a
            data stream

        Raises
        ------
        EmptyFileError
            dmap_file is empty
        FileNotFoundError
            dmap_file path does not exist

        See Also
        --------
        DmapRead : for inheritance information
        """
        DmapRead.__init__(self, filename, stream)

    # helper function that could be used parallelization
    def _read_darn_record(self, format_fields: List[dict]):
        """
        Read SuperDARN DMAP records from the DMAP byte array. Several SuperDARN
        field checks are done to insure the integrity of the file.

        Parameters
        ----------
        format_fields : List[dict]
            Is a list of dictionaries for the fields that are possible in a
            SuperDARN file type

        Raises:
        -------
        SuperDARNFieldMissingError - when a field is missing from the SuperDARN
                                file/stream type
        SuperDARNExtraFieldError - when an extra field is present in the
                                SuperDARN file/stream type
        SuperDARNDataFormatTypeError - when a field has the incorrect
                                field type for the SuperDARN file/stream type

        See Also
        --------
        read_record() - inherited from DmapRead
        missing_field_check(format_fields, record, self.rec_num) - checks
                        for missing fields. See this method for information
                        on why we use format_fields.
        extra_field_check(format_fields, record, self.rec_num) - checks for
                        extra fields in the record
        incorrect_types_check(format_fields, record, self.rec_num) - checks
                        for incorrect data types for SuperDARN file fields
        """
        record = self.read_record()
        SDarnUtilities.missing_field_check(format_fields, record, self.rec_num)
        SDarnUtilities.extra_field_check(format_fields, record, self.rec_num)
        SDarnUtilities.incorrect_types_check(format_fields, record,
                                             self.rec_num)
        self._dmap_records.append(record)

    def _read_darn_records(self, format_fields: List[dict]):
        """
        loops over the bytes in the in the SuperDARN byte array and
        calls the helper method read the SuperDARN records from the file/stream

        Parameters
        ----------
        format_fields : List[dict]
            Is a list of dictionaries for the fields that are possible in a
            SuperDARN file type. See missing_field_check method in
            SDarnUtilities for more information on this parameter

        See Also
        --------
        _read_darn_record
        """
        self.rec_num = 0  # record number, for exception info
        while self.cursor < self.dmap_end_bytes:
            self._read_darn_record(format_fields)
            self.rec_num += 1

    @property
    def get_dmap_records(self):
        """
        Gets the DMap records from the file read in
        """
        return self._dmap_records

    @property
    def get_records(self):
        """
        Gets the records containing the dictionary of fields from
        the file read in
        """
        return self.records

    def read_iqdat(self) -> List[dict]:
        """
        Reads iqdat DMAP file/stream

        Returns
        -------
        dmap_records : List[dict]
            DMAP record of the Iqdat data
        """
        pydarn_log.debug("Reading Iqdat file: {}".format(self.dmap_file))
        file_struct_list = [superdarn_formats.Iqdat.types]
        self._read_darn_records(file_struct_list)
        self.records = dmap2dict(self._dmap_records)
        return self.records

    def read_rawacf(self) -> List[dict]:
        """
        Reads Rawacf DMAP file/stream

        Returns
        -------
        dmap_records : List[dict]
            DMAP record of the Rawacf data
        """
        pydarn_log.debug("Reading Rawacf file: {}".format(self.dmap_file))

        file_struct_list = [superdarn_formats.Rawacf.types,
                            superdarn_formats.Rawacf.extra_fields,
                            superdarn_formats.Rawacf.cross_correlation_field]
        self._read_darn_records(file_struct_list)
        self.records = dmap2dict(self._dmap_records)
        return self.records

    def read_fitacf(self) -> List[dict]:
        """
        Reads Fitacf DMAP file/stream

        Returns
        -------
        dmap_records : List[dict]
            DMAP record of the Fitacf data
        """
        pydarn_log.debug("Reading Fitacf file: {}".format(self.dmap_file))

        # We need to separate the fields into subsets because fitacf fitting
        # methods 2.5 and 3.0 do not include a subset of fields if the data
        # quality is not "good". See missing_field_check method in
        # SDarnUtilities for more information.
        file_struct_list = [superdarn_formats.Fitacf.types,
                            superdarn_formats.Fitacf.extra_fields,
                            superdarn_formats.Fitacf.fitted_fields,
                            superdarn_formats.Fitacf.elevation_fields]
        self._read_darn_records(file_struct_list)
        self.records = dmap2dict(self._dmap_records)
        return self.records

    def read_grid(self) -> List[dict]:
        """
        Reads Grid DMAP file/stream

        Returns
        -------
        dmap_records : List[dict]
            DMAP record of the Grid data
        """
        pydarn_log.debug("Reading Grid file: {}".format(self.dmap_file))
        # We need to separate the fields into subsets because grid files
        # can exclude extra fields if the option -ext is not passed in
        # when generating the grid file in RST. See missing_field_check
        # method in SDarnUtilities for more information.
        file_struct_list = [superdarn_formats.Grid.types,
                            superdarn_formats.Grid.fitted_fields,
                            superdarn_formats.Grid.extra_fields]
        self._read_darn_records(file_struct_list)
        self.records = dmap2dict(self._dmap_records)
        return self.records

    def read_map(self) -> List[dict]:
        """
        Reads Map DMAP file/stream

        Returns
        -------
        dmap_records : List[dict]
            DMAP record of the Map data
        """
        pydarn_log.debug("Reading Map file: {}".format(self.dmap_file))
        # We need to separate the fields into subsets because map files
        # can exclude extra fields if the grid file does not contain them.
        # Other subsets are related to what processing commands are used to
        # generate the final map file, example: map_addhmb. This command is
        # not necessarily to generate a map file but it add the Heppner Maynard
        # boundary to the map file. See missing_field_check
        # method in SDarnUtilities for more information.
        file_struct_list = [superdarn_formats.Map.types,
                            superdarn_formats.Map.extra_fields,
                            superdarn_formats.Map.fit_fields,
                            superdarn_formats.Map.hmb_fields,
                            superdarn_formats.Map.model_fields]
        self._read_darn_records(file_struct_list)
        self.records = dmap2dict(self._dmap_records)
        return self.records


class SDarnWrite(DmapWrite):
    """
    Writes Dmap records to file or stream and writes SuperDARN file format.
    ...

    Attributes
    -----------
    dmap_records : List[dict]
        List of dmap records
    filename : str
        Name of the file the user wants to write to
    dmap_bytearr : bytearray
        Byte array representing the dmap records in bytes

    Methods
    -------
    write_iqdat(filename)
        Writes dmap records to SuperDARN IQDAT file structure
        with the given filename
    write_fitacf(filename)
        Write dmap records to SuperDARN RAWACF file structure
        with the given filename
    write_rawacf(filename)
        Writes dmap records to SuperDARN FITACF file structure
        with the given filename
    write_grid(filename)
        Writes dmap records to SuperDARN GRID file structure
        with the given filename
    write_map(filename)
        Writes dmap records to SuperDARN MAP file structure
        with the given filename
    SuperDARN_file_structure_to_bytes(file_struct_list)
        Converts dmap records to SuperDARN file structure bytes based
        on file_struct_list

    See Also
    --------
    DmapWrite
        class SDarnWrite inherits from
    dmap_scalar_to_bytes(scalar)
        Converts a DmapScalar to bytes
    dmap_array_to_bytes(array)
        Converts a DmapArray to bytes
    """

    def __init__(self, dmap_records: List[dict] = [], filename: str = ""):
        """
        Inherits from DmapWrite class and checks the data record is not empty.
        This check will be taken out when real-time is implemented.

        Parameters
        ----------
        dmap_records : List[dict]
            List of dmap records
        filename : str
            Name of the file the user wants to write to
        """
        DmapWrite.__init__(self, dmap_records, filename)

        # WARNING: This check will be removed when real-time is implemented to
        # allow for constantly updating data from a processing data feed
        self._empty_record_check()

    def write_iqdat(self, filename: str = ""):
        """
        Writes SuperDARN file type IQDAT

        Parameters:
        -----------
        filename : str
            The name of the IQDAT file including path


        Raises:
        -------
        superDARNExtraFieldError - if there is an extra field
        SuperDARNFieldMissingError- if there is an missing field
        SuperDARNDataFormatTypeError - if there is a formatting error
                               like an incorrect data type format

        See Also:
        ---------
        extra_field_check
        missing_field_check
        superdarn_formats.Iqdat - module contain the data types
                                 in each SuperDARN files types
        """
        self._filename_check(filename)
        self._empty_record_check()
        pydarn_log.debug("Writing Iqdat file: {}".format(self.filename))

        file_struct_list = [superdarn_formats.Iqdat.types]
        self.superDARN_file_structure_to_bytes(file_struct_list)
        with open(self.filename, 'wb') as f:
            f.write(self.dmap_bytearr)

    def write_rawacf(self, filename: str = ""):
        """
        Writes SuperDARN file type RAWACF

        Parameters:
        -----------
        filename : str
            The name of the RAWACF file including path


        Raises:
        -------
        superDARNExtraFieldError - if there is an extra field
        SuperDARNFieldMissingError- if there is an missing field
        SuperDARNDataFormatTypeError - if there is a formatting error
                               like an incorrect data type format

        See Also:
        ---------
        extra_field_check
        missing_field_check
        superdarn_formats.Rawacf - module contain the data types
                                 in each SuperDARN files types
        """
        pydarn_log.debug("Writing Rawacf file: {}".format(self.filename))
        self._filename_check(filename)
        self._empty_record_check()
        file_struct_list = [superdarn_formats.Rawacf.types,
                            superdarn_formats.Rawacf.extra_fields,
                            superdarn_formats.Rawacf.cross_correlation_field]
        self.superDARN_file_structure_to_bytes(file_struct_list)
        with open(self.filename, 'wb') as f:
            f.write(self.dmap_bytearr)

    def write_fitacf(self, filename: str = ""):
        """
        Writes SuperDARN file type FITACF

        Parameters:
        -----------
        filename : str
            The name of the FITACF file including path


        Raises:
        -------
        superDARNExtraFieldError - if there is an extra field
        SuperDARNFieldMissingError- if there is an missing field
        SuperDARNDataFormatTypeError - if there is a formatting error
                               like an incorrect data type format

        See Also:
        ---------
        extra_field_check
        missing_field_check
        superdarn_formats.Fitacf - module contain the data types
                                 in each SuperDARN files types
        """
        pydarn_log.debug("Writing Fitacf file: {}".format(self.filename))

        self._filename_check(filename)
        self._empty_record_check()
        # We need to separate the fields into subsets because fitacf fitting
        # methods 2.5 and 3.0 do not include a subset of fields if the data
        # quality is not "good". See missing_field_check method in
        # SDarnUtilities for more information.
        file_struct_list = [superdarn_formats.Fitacf.types,
                            superdarn_formats.Fitacf.extra_fields,
                            superdarn_formats.Fitacf.fitted_fields,
                            superdarn_formats.Fitacf.elevation_fields]
        self.superDARN_file_structure_to_bytes(file_struct_list)
        with open(self.filename, 'wb') as f:
            f.write(self.dmap_bytearr)

    def write_grid(self, filename: str = ""):
        """
        Writes SuperDARN file type GRID

        Parameters:
        -----------
        filename : str
            The name of the GRID file including path


        Raises:
        -------
        superDARNExtraFieldError - if there is an extra field
        SuperDARNFieldMissingError- if there is an missing field
        SuperDARNDataFormatTypeError - if there is a formatting error
                               like an incorrect data type format

        See Also:
        ---------
        extra_field_check
        missing_field_check
        superdarn_formats.Grid - module contain the data types
                                 in each SuperDARN files types
        """
        pydarn_log.debug("Writing Grid file: {}".format(self.filename))

        self._filename_check(filename)
        self._empty_record_check()
        # We need to separate the fields into subsets because grid files
        # can exclude extra fields if the option -ext is not passed in
        # when generating the grid file in RST. See missing_field_check
        # method in SDarnUtilities for more information.
        file_struct_list = [superdarn_formats.Grid.types,
                            superdarn_formats.Grid.fitted_fields,
                            superdarn_formats.Grid.extra_fields]
        self.superDARN_file_structure_to_bytes(file_struct_list)
        with open(self.filename, 'wb') as f:
            f.write(self.dmap_bytearr)

    def write_map(self, filename: str = ""):
        """
        Writes SuperDARN file type MAP

        Parameters:
        -----------
        filename : str
            The name of the MAP file including path


        Raises:
        -------
        superDARNExtraFieldError - if there is an extra field
        SuperDARNFieldMissingError- if there is an missing field
        SuperDARNDataFormatTypeError - if there is a formatting error
                               like an incorrect data type format

        See Also:
        ---------
        extra_field_check
        missing_field_check
        superdarn_formats.Map - module contain the data types
                                 in each SuperDARN files types
        """
        pydarn_log.debug("Writing Map file: {}".format(self.filename))

        self._filename_check(filename)
        self._empty_record_check()
        # We need to separate the fields into subsets because map files
        # can exclude extra fields if the grid file does not contain them.
        # Other subsets are related to what processing commands are used to
        # generate the final map file, example: map_addhmb. This command is
        # not necessarily to generate a map file but it add the Heppner Maynard
        # boundary to the map file. See missing_field_check
        # method in SDarnUtilities for more information.
        file_struct_list = [superdarn_formats.Map.types,
                            superdarn_formats.Map.extra_fields,
                            superdarn_formats.Map.fit_fields,
                            superdarn_formats.Map.model_fields,
                            superdarn_formats.Map.hmb_fields]
        self.superDARN_file_structure_to_bytes(file_struct_list)
        with open(self.filename, 'wb') as f:
            f.write(self.dmap_bytearr)

    def superDARN_file_structure_to_bytes(self, file_struct_list: List[dict]):
        """
        Checks the DMAP records are the correct structure of the file type and
        then uses the DmapWrite writing method to covert the record to bytes.

        Parameters
        ----------
        file_struct_list : List[dict]
        A list of possible fields for the given SuperDARN file type

        Raises
        ------
        SuperDARNFieldMissingError - Missing fields in the DMAP record that is
        required for the SuperDARN file type
        SuperDARNExtraFieldError - Extra fields in the DMAP record that are not
        allowed in the SuperDARN file type
        SuperDARNDataFormatTypeError - Wrong format data type error for the
        SuperDARN file type fields in the DMAP record

        See Also
        --------
        Within superdarn.py module:
        SDarnUtilities.missing_field_check(format_fields, record, self.rec_num)
                - checks for missing fields. See this method for information
                  on why we use format_fields.
        DarnUtilties.extra_field_check(format_fields, record, self.rec_num)
                - checks for extra fields in the record
        SDarnUtilities.incorrect_types_check(format_fields, record,
                                            self.rec_num)
                - checks for incorrect data types for SuperDARN file fields

        """
        self.rec_num = 0
        pydarn_log.debug("Checking and converting SuperDARN data to bytes ")

        for self.rec_num in range(len(self.dmap_records)):
            record = self.dmap_records[self.rec_num]
            # field checks
            SDarnUtilities.extra_field_check(file_struct_list, record,
                                             self.rec_num)
            SDarnUtilities.missing_field_check(file_struct_list, record,
                                               self.rec_num)
            SDarnUtilities.incorrect_types_check(file_struct_list, record,
                                                 self.rec_num)
            # start converting
            self._dmap_record_to_bytes(record)



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marci Detwiller
"""
This file contains classes and functions for
converting of Borealis file types.

Classes
-------
BorealisConvert: Writes Borealis SuperDARN files types to
    SuperDARN SDARN files with DMap record structure

Exceptions
----------
BorealisFileTypeError
BorealisFieldMissingError
BorealisExtraFieldError
BorealisDataFormatTypeError
BorealisConversionTypesError
BorealisConvert2IqdatError
BorealisConvert2RawacfError

Notes
-----
BorealisConvert makes use of SDARNWrite to write to SuperDARN file types

See Also
--------
BorealisRead
BorealisWrite
BorealisSiteRead
BorealisSiteWrite
BorealisArrayRead
BorealisArrayWrite

For more information on Borealis data files and how they convert to dmap,
see: https://borealis.readthedocs.io/en/latest/

Future Work
-----------
Remove requirement for hdf5 filename by fixing Borealis datawrite so that
    slice_id is written. This must be fixed in the Borealis code.
Update noise values in SDarn fields when these can be calculated.

"""
import logging
import math
import numpy as np

from datetime import datetime
from typing import Union

from pydarn import (borealis_exceptions, BorealisRead, SDarnWrite, dict2dmap)

pydarn_log = logging.getLogger('pydarn')

# 3 letter radar code, mapped to station id for SDarn files conversion.
# TODO: when merged with plotting, remove this dictionary and call the
#    one in the plotting folder... also move Radars.py to a more
#    central location.
code_to_stid = {
    "tst": 0,
    "gbr": 1,
    "sch": 2,
    "kap": 3,
    "hal": 4,
    "sas": 5,
    "pgr": 6,
    "kod": 7,
    "sto": 8,
    "pyk": 9,
    "han": 10,
    "san": 11,
    "sys": 12,
    "sye": 13,
    "tig": 14,
    "ker": 15,
    "ksr": 16,
    "unw": 18,
    "zho": 19,
    "mcm": 20,
    "fir": 21,
    "sps": 22,
    "bpk": 24,
    "wal": 32,
    "bks": 33,
    "hok": 40,
    "hkw": 41,
    "inv": 64,
    "rkn": 65,
    "cly": 66,
    "dce": 96,
    "svb": 128,
    "fhw": 204,
    "fhe": 205,
    "cvw": 206,
    "cve": 207,
    "adw": 208,
    "ade": 209,
    "ekb": 512,
}


class BorealisConvert(BorealisRead):
    """
    Class for converting Borealis filetypes to SDARN DMap filetypes.

    See Also
    --------
    BorealisRawacf
    BorealisBfiq
    Rawacf
    Iqdat
    BorealisRead
    BorealisSiteRead
    BorealisArrayRead
    SDarnWrite

    Attributes
    ----------
    allowed_conversions: dict
        Mapping of allowed Borealis types to their corresponding
        SDARN DMap converted types.
    filename: str; borealis_filename: str
        The filename of the Borealis HDF5 file being read.
    borealis_filetype: str
        The type of Borealis file. Types include:
        'bfiq'
        'rawacf'
    reader: Union[BorealisSiteWrite, BorealisArrayWrite]
        the wrapped BorealisSiteRead or BorealisArrayRead instance
    borealis_file_structure: Union[str, None]
        The structure of the file read, 'site' or 'array'. Default None.
    records: OrderedDict{dict}; borealis_records: OrderedDict{dict}
        The Borealis data in a dictionary of records, according to the
        site file format. The conversion functions use
        record format only as that is what is used by SDARN DMap
        files.
    arrays: dict
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
    record_names: list[str]
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms since
        epoch), and are equal to the group names in the site file types.
    sdarn_filename: str
        The filename of the SDARN DMap file to be written.
    sdarn_dmap_records: list[dict]
        The converted DMap records to write to file.
    sdarn_dict: dict
        The dictionary of SDARN records before the conversion to DMap format.
    sdarn_filetype: str
        The dmap filetype converted to. 'rawacf' and 'iqdat' are allowed.
    borealis_slice_id: int
        The slice id of the file being converted. Used as channel identifier
        in SDARN DMap records.
    """

    __allowed_conversions = {'rawacf': 'rawacf', 'bfiq': 'iqdat'}

    def __init__(self, borealis_filename: str, borealis_filetype: str,
                 sdarn_filename: str, borealis_slice_id: int,
                 borealis_file_structure: Union[str, None] = None):
        """
        Convert HDF5 Borealis records to a given SDARN file with DMap format.

        Parameters
        ----------
        borealis_filename: str
            file name containing Borealis hdf5 data. Name contains slice id
            and allows the file read.
        borealis_filetype: str
            The type of Borealis file. Types allowed to convert:
            'bfiq' -> 'iqdat'
            'rawacf' -> 'rawacf'
        sdarn_filename: str
            The filename of the SDARN DMap file to be written.
        borealis_slice_id: int
            The slice id of the file being converted.
        borealis_file_structure: Union[str, None]
            The write structure of the file provided. Possible types are
            'site', 'array', or None. If None (default), array read will be
            attempted first followed by site.

        Raises
        ------
        BorealisConversionTypesError
        ConvertFileOverWriteError
        """

        super(BorealisConvert, self).__init__(borealis_filename,
                                              borealis_filetype,
                                              borealis_file_structure)

        # Attributes written in the parent init (BorealisRead):
        #   filename : str
        #   borealis_filetype : str
        #   reader: Union[BorealisSiteRead, BorealisArrayRead]
        #   borealis_file_structure : str
        #   records : dict (of data)
        #   arrays : dict (of data)

        self.borealis_records = self.records
        self.sdarn_filename = sdarn_filename
        self.borealis_filename = self.filename
        self._borealis_slice_id = borealis_slice_id
        self._sdarn_dmap_records = {}
        self._sdarn_dict = {}
        try:
            self._sdarn_filetype = self.__allowed_conversions[
                    self.borealis_filetype]
        except KeyError:
            raise borealis_exceptions.BorealisConversionTypesError(
                self.sdarn_filename, self.borealis_filetype,
                self.__allowed_conversions)

        if self.borealis_filename == self.sdarn_filename:
            raise borealis_exceptions.ConvertFileOverWriteError(
                    self.borealis_filename)
        self._write_to_sdarn()

    def __repr__(self):
        """ for representation of the class object"""

        return "{class_name}({records}{borealis_filetype}{sdarn_filename})"\
               "".format(class_name=self.__class__.__name__,
                         records=self.records,
                         borealis_filetype=self.borealis_filetype,
                         sdarn_filename=self.sdarn_filename)

    def __str__(self):
        """ for printing of the class object"""

        return "Converting {total_records} {borealis_filetype} records into "\
               "DMap SDARN records and writing to file {sdarn_filename}."\
               "".format(total_records=len(self.borealis_records.keys()),
                         borealis_filetype=self.borealis_filetype,
                         sdarn_filename=self.sdarn_filename)

    @property
    def sdarn_dmap_records(self):
        """
        The converted SDARN DMap records to write to file.
        """
        return self._sdarn_dmap_records

    @property
    def sdarn_dict(self):
        """
        The converted SDARN records as a dictionary, before being converted
        to DMap.
        """
        return self._sdarn_dict

    @property
    def sdarn_filetype(self):
        """
        The dmap filetype converted to. 'rawacf' and 'iqdat' are allowed.
        """
        return self._sdarn_filetype

    @property
    def borealis_slice_id(self):
        """
        The slice id of the file being converted. Used as channel identifier
        in SDARN DMap records.
        """
        return self._borealis_slice_id

    def _write_to_sdarn(self) -> str:
        """
        Write the Borealis records as SDARN DMap records to a file using PyDARN
        IO.

        Returns
        -------
        sdarn_filename, the name of the SDARN file written.
        """

        self._convert_records_to_dmap()
        sdarn_writer = SDarnWrite(self._sdarn_dmap_records,
                                  self.sdarn_filename)
        if self.sdarn_filetype == 'iqdat':
            sdarn_writer.write_iqdat(self.sdarn_filename)
        elif self.sdarn_filetype == 'rawacf':
            sdarn_writer.write_rawacf(self.sdarn_filename)
        return self.sdarn_filename

    def _convert_records_to_dmap(self):
        """
        Convert the Borealis records to the a DMap filetype according to
        the origin filetype.

        Raises
        ------
        BorealisConversionTypesError
        """
        if self.sdarn_filetype == 'iqdat':
            if self._is_convertible_to_iqdat():
                self._convert_bfiq_to_iqdat()
        elif self.sdarn_filetype == 'rawacf':
            if self._is_convertible_to_rawacf():
                self._convert_rawacf_to_rawacf()
        else:  # nothing else is currently supported
            raise borealis_exceptions.BorealisConversionTypesError(
                self.sdarn_filename, self.borealis_filetype,
                self.__allowed_conversions)

    def _is_convertible_to_iqdat(self) -> bool:
        """
        Checks if the file is convertible to iqdat.

        The file is convertible if:
            - the origin filetype is bfiq
            - the blanked_samples array = pulses array for all records
            - the pulse_phase_offset array contains all zeroes for all records

        Raises
        ------
        BorealisConversionTypesError
        BorealisConvert2IqdatError

        Returns
        -------
        True if convertible
        """
        if self.borealis_filetype != 'bfiq':
            raise borealis_exceptions.BorealisConversionTypesError(
                self.sdarn_filename, self.borealis_filetype,
                self.__allowed_conversions)
        else:  # There are some specific things to check
            for k, v in self.borealis_records.items():
                if not np.array_equal(v['blanked_samples'],
                                      v['pulses'] *
                                      int(v['tau_spacing']/v['tx_pulse_len'])):
                    raise borealis_exceptions.\
                            BorealisConvert2IqdatError('Increased complexity:'
                                                       ' Borealis bfiq file'
                                                       ' record {}'
                                                       ' blanked_samples {}'
                                                       ' does not equal pulses'
                                                       ' array {}'
                                                       ''.format(k,
                                                                 v['blanked_samples'],
                                                                 v['pulses']))
                if not all([x == 0 for x in v['pulse_phase_offset']]):
                    raise borealis_exceptions.\
                            BorealisConvert2IqdatError('Increased complexity:'
                                                       ' Borealis bfiq file'
                                                       ' record {}'
                                                       ' pulse_phase_offset {}'
                                                       ' contains non-zero'
                                                       ' values.'
                                                       ''.format(k,
                                                                 v['pulse_phase_offset']))
        return True

    def _is_convertible_to_rawacf(self) -> bool:
        """
        Checks if the file is convertible to rawacf.

        The file is convertible if:
            - the origin filetype is rawacf
            - the blanked_samples array = pulses array for all records
            - the pulse_phase_offset array contains all zeroes for all records

        TODO: should this fail for multiple beams in the same
        integration time. IE, is it ok for dmap files to have multiple
        records with same origin time and timestamps due to a different
        beam azimuth.

        Raises
        ------
        BorealisConversionTypesError
        BorealisConvert2RawacfError

        Returns
        -------
        True if convertible
        """
        if self.borealis_filetype != 'rawacf':
            raise borealis_exceptions.\
                    BorealisConversionTypesError(self.sdarn_filename,
                                                 self.borealis_filetype,
                                                 self.__allowed_conversions)
        else:  # There are some specific things to check
            for k, v in self.borealis_records.items():
                if not np.array_equal(v['blanked_samples'],
                                      v['pulses'] *
                                      int(v['tau_spacing']/v['tx_pulse_len'])):
                    raise borealis_exceptions.\
                            BorealisConvert2RawacfError('Increased complexity:'
                                                        ' Borealis rawacf file'
                                                        ' record {}'
                                                        ' blanked_samples {}'
                                                        'does not equal'
                                                        ' pulses array {}'
                                                        ''.format(k,
                                                                  v['blanked_samples'],
                                                                  v['pulses']))

        return True

    def _convert_bfiq_to_iqdat(self):
        """
        Conversion for bfiq to iqdat SDARN DMap records.

        See Also
        --------
        https://superdarn.github.io/rst/superdarn/src.doc/rfc/0027.html
        https://borealis.readthedocs.io/en/latest/
        BorealisBfiq
        Iqdat

        Raises
        ------
        BorealisConvert2IqdatError

        Returns
        -------
        dmap_recs, the records converted to DMap format
        """
        try:
            recs = []
            for record in self.borealis_records.items():
                record_dict = \
                        self.__convert_bfiq_record(self.borealis_slice_id,
                                                   record,
                                                   self.borealis_filename)
                recs.append(record_dict)
            self._sdarn_dict = recs
            self._sdarn_dmap_records = dict2dmap(recs)
        except Exception as e:
            raise borealis_exceptions.BorealisConvert2IqdatError(e) from e

    @staticmethod
    def __convert_bfiq_record(borealis_slice_id: int,
                              borealis_bfiq_record: tuple,
                              origin_string: str) -> dict:
        """
        Converts a single record dict of Borealis bfiq data to a SDARN DMap
        record dict.

        Parameters
        ----------
        borealis_slice_id : int
            slice id integer of the borealis data, for conversion. Used
            as SDARN DMap channel identifier.
        borealis_bfiq_record : tuple(str, dict)
            Key is bfiq record timestamp, value is dictionary of bfiq
            record data.
        origin_string : str
            String representing origin of the Borealis data, typically
            Borealis filename.
        """

        # key value pair from Borealis bfiq record.
        (k, v) = borealis_bfiq_record

        # data_descriptors (dimensions) are num_antenna_arrays,
        # num_sequences, num_beams, num_samps
        # scale by normalization and then scale to integer max as per
        # dmap style
        data = v['data'].reshape(v['data_dimensions']).astype(np.complex128) / \
            v['data_normalization_factor'] * np.iinfo(np.int16).max

        # Borealis git tag version numbers. If not a tagged version,
        # then use 255.255
        if v['borealis_git_hash'][0] == 'v' and \
                v['borealis_git_hash'][2] == '.':

            borealis_major_revision = v['borealis_git_hash'][1]
            borealis_minor_revision = v['borealis_git_hash'][3]
        else:
            borealis_major_revision = 255
            borealis_minor_revision = 255

        # base offset for setting the toff field in SDARN DMap iqdat file.
        offset = 2 * v['antenna_arrays_order'].shape[0] * v['num_samps']

        for beam_index, beam in enumerate(v['beam_nums']):
            # grab this beam's data
            # shape is now num_antenna_arrays x num_sequences
            # x num_samps
            this_data = data[:, :, beam_index, :]
            # iqdat shape is num_sequences x num_antennas_arrays x
            # num_samps x 2 (real, imag), flattened
            reshaped_data = []
            for i in range(v['num_sequences']):
                # get the samples for each array 1 after the other
                arrays = [this_data[x, i, :]
                          for x in range(this_data.shape[0])]
                # append
                reshaped_data.append(np.ravel(np.column_stack(arrays)))

            # (num_sequences x num_antenna_arrays x num_samps,
            # flattened)
            flattened_data = np.array(reshaped_data).flatten()

            int_data = np.empty(flattened_data.size * 2, dtype=np.int16)
            int_data[0::2] = flattened_data.real
            int_data[1::2] = flattened_data.imag

            # flattening done in convert_to_dmap_datastructures
            record_dict = {
                'radar.revision.major': np.int8(borealis_major_revision),
                'radar.revision.minor': np.int8(borealis_minor_revision),
                'origin.code': np.int8(100),  # indicating Borealis
                'origin.time': datetime.\
                utcfromtimestamp(v['sqn_timestamps'][0]).strftime("%c"),
                'origin.command': 'Borealis ' + v['borealis_git_hash'] + \
                ' ' + v['experiment_name'],
                'cp': np.int16(v['experiment_id']),
                'stid': np.int16(code_to_stid[v['station']]),
                'time.yr': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    year),
                'time.mo': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    month),
                'time.dy': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    day),
                'time.hr': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    hour),
                'time.mt': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    minute),
                'time.sc': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    second),
                'time.us': np.int32(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    microsecond),
                'txpow': np.int16(-1),
                'nave': np.int16(v['num_sequences']),
                'atten': np.int16(0),
                'lagfr': np.int16(v['first_range_rtt']),
                # smsep is in us; conversion from seconds
                'smsep': np.int16(1e6 / v['rx_sample_rate']),
                'ercod': np.int16(0),
                # TODO: currently not implemented
                'stat.agc': np.int16(0),
                # TODO: currently not implemented
                'stat.lopwr': np.int16(0),
                # TODO: currently not implemented
                'noise.search': np.float32(v['noise_at_freq'][0]),
                # TODO: currently not implemented
                'noise.mean': np.float32(0),
                'channel': np.int16(borealis_slice_id),
                'bmnum': np.int16(beam),
                'bmazm': np.float32(v['beam_azms'][beam_index]),
                'scan': np.int16(v['scan_start_marker']),
                # no digital receiver offset or rxrise required in
                # Borealis
                'offset': np.int16(0),
                'rxrise': np.int16(0),
                'intt.sc': np.int16(math.floor(v['int_time'])),
                'intt.us': np.int32(math.fmod(v['int_time'], 1.0) * 1e6),
                'txpl': np.int16(v['tx_pulse_len']),
                'mpinc': np.int16(v['tau_spacing']),
                'mppul': np.int16(len(v['pulses'])),
                # an alternate lag-zero will be given, so subtract 1.
                'mplgs': np.int16(v['lags'].shape[0] - 1),
                'nrang': np.int16(v['num_ranges']),
                'frang': np.int16(round(v['first_range'])),
                'rsep': np.int16(round(v['range_sep'])),
                'xcf': np.int16('intf' in v['antenna_arrays_order']),
                'tfreq': np.int16(v['freq']),
                # mxpwr filler; cannot specify this information
                'mxpwr': np.int32(-1),
                # lvmax RST default
                'lvmax': np.int32(20000),
                'iqdata.revision.major': np.int32(1),
                'iqdata.revision.minor': np.int32(0),
                'combf': 'Converted from Borealis file: ' + origin_string +\
                         ' record ' + str(k) + \
                         ' ; Number of beams in record: ' + \
                         str(len(v['beam_nums'])) + ' ; ' + \
                         v['experiment_comment'] + ' ; ' + \
                         v['slice_comment'],
                'seqnum': np.int32(v['num_sequences']),
                'chnnum': np.int32(v['antenna_arrays_order'].shape[0]),
                'smpnum': np.int32(v['num_samps']),
                # NOTE: The following is a hack. This is currently how
                # iqdat files are being processed . RST make_raw does
                # not use first range information at all, only skip
                # number.
                # However ROS provides the number of ranges to the
                # first range as the skip number. Skip number is
                # documented as number to identify bad ranges due
                # to digital receiver rise time. Borealis skpnum should
                # in theory =0 as the first sample from Borealis
                # decimated (prebfiq) data is centred on the first
                # pulse.
                'skpnum': np.int32(v['first_range'] / v['range_sep']),
                'ptab': v['pulses'].astype(np.int16),
                'ltab': v['lags'].astype(np.int16),
                # timestamps in ms, convert to seconds and us.
                'tsc': np.array([math.floor(x/1e3) for x in
                                 v['sqn_timestamps']], dtype=np.int32),
                'tus': np.array([math.fmod(x, 1000.0) * 1e3 for x in
                                 v['sqn_timestamps']], dtype=np.int32),
                'tatten': np.array([0] * v['num_sequences'],
                                   dtype=np.int16),
                'tnoise': v['noise_at_freq'].astype(np.float32),
                'toff': np.array([i * offset for i in
                                  range(v['num_sequences'])], dtype=np.int32),
                'tsze': np.array([offset] * v['num_sequences'],
                                 dtype=np.int32),
                'data': int_data
            }
        return record_dict

    def _convert_rawacf_to_rawacf(self):
        """
        Conversion for Borealis hdf5 rawacf to SDARN DMap rawacf files.

        See Also
        --------
        __convert_rawacf_record
        https://superdarn.github.io/rst/superdarn/src.doc/rfc/0008.html
        https://borealis.readthedocs.io/en/latest/
        BorealisRawacf
        Rawacf

        Raises
        ------
        BorealisConvert2RawacfError

        Returns
        -------
        dmap_recs, the records converted to DMap format
        """
        try:
            recs = []
            for record in self.borealis_records.items():
                record_dict = \
                        self.__convert_rawacf_record(self.borealis_slice_id,
                                                     record,
                                                     self.borealis_filename)
                recs.append(record_dict)
            self._sdarn_dict = recs
            self._sdarn_dmap_records = dict2dmap(recs)
        except Exception as e:
            raise borealis_exceptions.BorealisConvert2RawacfError(e) from e

    @staticmethod
    def __convert_rawacf_record(borealis_slice_id: int,
                                borealis_rawacf_record: tuple,
                                origin_string: str) -> dict:
        """
        Converts a single record dict of Borealis rawacf data to a SDARN DMap
        record dict.

        Parameters
        ----------
        borealis_slice_id : int
            slice id integer of the borealis data, for conversion. Used
            as SDARN DMap channel identifier.
        borealis_rawacf_record : tuple(str, dict)
            Key is rawacf record timestamp, value is dictionary of rawacf
            record data.
        origin_string : str
            String representing origin of the Borealis data, typically
            Borealis filename.
        """

        # key value pair from Borealis record dictionary
        (k, v) = borealis_rawacf_record

        shaped_data = {}
        # correlation_descriptors are num_beams, num_ranges, num_lags
        # scale by the scale squared to make up for the multiply
        # in correlation (integer max squared)
        shaped_data['main_acfs'] = v['main_acfs'].reshape(
            v['correlation_dimensions']).astype(
            np.complex128) * ((np.iinfo(np.int16).max**2) /
                              (v['data_normalization_factor']**2))

        if 'intf_acfs' in v.keys():
            shaped_data['intf_acfs'] = v['intf_acfs'].reshape(
                v['correlation_dimensions']).astype(np.complex128) * \
                    ((np.iinfo(np.int16).max**2) /
                     (v['data_normalization_factor']**2))
        if 'xcfs' in v.keys():
            shaped_data['xcfs'] = v['xcfs'].reshape(
                v['correlation_dimensions']).astype(np.complex128) *\
                    ((np.iinfo(np.int16).max**2) /
                     (v['data_normalization_factor']**2))

        # Borealis git tag version numbers. If not a tagged version,
        # then use 255.255
        if v['borealis_git_hash'][0] == 'v' and \
                v['borealis_git_hash'][2] == '.':
            borealis_major_revision = v['borealis_git_hash'][1]
            borealis_minor_revision = v['borealis_git_hash'][3]
        else:
            borealis_major_revision = 255
            borealis_minor_revision = 255

        for beam_index, beam in enumerate(v['beam_nums']):
            # this beam, all ranges lag 0
            lag_zero = shaped_data['main_acfs'][beam_index, :, 0]
            lag_zero[-10:] = shaped_data['main_acfs'][beam_index, -10:, -1]
            lag_zero_power = (lag_zero.real**2 + lag_zero.imag**2)**0.5

            correlation_dict = {}
            for key in shaped_data:
                # num_ranges x num_lags (complex)
                this_correlation = shaped_data[key][beam_index, :, :-1]
                # set the lag0 to the alternate lag0 for the end of the
                # array (when interference of first pulse would occur)
                this_correlation[-10:, 0] = \
                    shaped_data[key][beam_index, -10:, -1]
                # shape num_beams x num_ranges x num_la gs, now
                # num_ranges x num_lags-1 b/c alternate lag-0 combined
                # with lag-0 (only used for last ranges)

                # (num_ranges x num_lags, flattened)
                flattened_data = np.array(this_correlation).flatten()

                int_data = np.empty(flattened_data.size * 2, dtype=np.float32)
                int_data[0::2] = flattened_data.real
                int_data[1::2] = flattened_data.imag
                # num_ranges x num_lags x 2; num_lags is one less than
                # in Borealis file because Borealis keeps alternate
                # lag0
                new_data = int_data.reshape(v['correlation_dimensions'][1],
                                            v['correlation_dimensions'][2]-1,
                                            2)
                # NOTE: Flattening happening in
                # convert_to_dmap_datastructures
                # place the SDARN-style array in the dict
                correlation_dict[key] = new_data

            record_dict = {
                'radar.revision.major': np.int8(borealis_major_revision),
                'radar.revision.minor': np.int8(borealis_minor_revision),
                'origin.code': np.int8(100),  # indicating Borealis
                'origin.time': datetime.\
                utcfromtimestamp(v['sqn_timestamps'][0]).strftime("%c"),
                'origin.command': 'Borealis ' + v['borealis_git_hash'] +\
                ' ' + v['experiment_name'],
                'cp': np.int16(v['experiment_id']),
                'stid': np.int16(code_to_stid[v['station']]),
                'time.yr': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    year),
                'time.mo': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    month),
                'time.dy': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    day),
                'time.hr': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    hour),
                'time.mt': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    minute),
                'time.sc': np.int16(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    second),
                'time.us': np.int32(datetime.
                                    utcfromtimestamp(v['sqn_timestamps'][0]).
                                    microsecond),
                'txpow': np.int16(-1),
                # see Borealis documentation
                'nave': np.int16(v['num_sequences']),
                'atten': np.int16(0),
                'lagfr': np.int16(v['first_range_rtt']),
                'smsep': np.int16(1e6/v['rx_sample_rate']),
                'ercod': np.int16(0),
                # TODO: currently not implemented
                'stat.agc': np.int16(0),
                # TODO: currently not implemented
                'stat.lopwr': np.int16(0),
                # TODO: currently not implemented
                'noise.search': np.float32(v['noise_at_freq'][0]),
                # TODO: currently not implemented
                'noise.mean': np.float32(0),
                'channel': np.int16(borealis_slice_id),
                'bmnum': np.int16(beam),
                'bmazm': np.float32(v['beam_azms'][beam_index]),
                'scan': np.int16(v['scan_start_marker']),
                # no digital receiver offset or rxrise required in
                # Borealis
                'offset': np.int16(0),
                'rxrise': np.int16(0),
                'intt.sc': np.int16(math.floor(v['int_time'])),
                'intt.us': np.int32(math.fmod(v['int_time'], 1.0) * 1e6),
                'txpl': np.int16(v['tx_pulse_len']),
                'mpinc': np.int16(v['tau_spacing']),
                'mppul': np.int16(len(v['pulses'])),
                # an alternate lag-zero will be given.
                'mplgs': np.int16(v['lags'].shape[0] - 1),
                'nrang': np.int16(v['correlation_dimensions'][1]),
                'frang': np.int16(round(v['first_range'])),
                'rsep': np.int16(round(v['range_sep'])),
                # False if list is empty.
                'xcf': np.int16(bool('xcfs' in v.keys())),
                'tfreq': np.int16(v['freq']),
                'mxpwr': np.int32(-1),
                'lvmax': np.int32(20000),
                'rawacf.revision.major': np.int32(1),
                'rawacf.revision.minor': np.int32(0),
                'combf': 'Converted from Borealis file: ' + origin_string + \
                         ' record ' + str(k) + \
                         ' ; Number of beams in record: ' + \
                         str(len(v['beam_nums'])) + ' ; ' + \
                         v['experiment_comment'] + ' ; ' + v['slice_comment'],
                'thr': np.float32(0),
                'ptab': v['pulses'].astype(np.int16),
                'ltab': v['lags'].astype(np.int16),
                'pwr0': lag_zero_power.astype(np.float32),
                # list from 0 to num_ranges
                'slist': np.array(list(range(0,
                                             v['correlation_dimensions'][1]))).astype(np.int16),
                'acfd': correlation_dict['main_acfs'],
                'xcfd': correlation_dict['xcfs']
            }

        return record_dict



# Copyright (C) 2019 SuperDARN
# Author: Marian Schmidt
# This code is improvement based on rti.py in the DaVitpy library
# https://github.com/vtsuperdarn/davitpy/blob/master/davitpy

"""
Range-Time Parameter (aka Intensity) plots
"""
import matplotlib.pyplot as plt
import numpy as np
import warnings

from datetime import datetime, timedelta
from matplotlib import dates, colors, cm, ticker
from typing import List

from pydarn import (dmap2dict, DmapArray, DmapScalar,
                    rtp_exceptions, SuperDARNCpids, SuperDARNRadars,
                    standard_warning_format, PyDARNColormaps)

warnings.formatwarning = standard_warning_format


class RTP():
    """
    Range-Time Parameter plots SuperDARN data using the following fields:

    Class pattern design: Builder
    This class inherits matplotlib.pyplot to inherit plotting features as well
    build off their builder design pattern.

    Notes
    -----
    This class inherits from matplotlib to generate the figures

    Methods
    -------
    plot_profile
    plot_scalar
    plot_summary
    """

    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_rang_time()\n"\
                "   - plot_time_series()\n"\
                "   - plot_summary()\n"

    @classmethod
    def plot_range_time(cls, dmap_data: List[dict], parameter: str = 'v',
                        beam_num: int = 0, channel: int = 'all', ax=None,
                        background: str = 'w', groundscatter: bool = False,
                        zmin: int = None, zmax: int = None,
                        start_time: datetime = None, end_time: datetime = None,
                        colorbar: plt.colorbar = None, ymax: int = None,
                        colorbar_label: str = '', norm=colors.Normalize,
                        cmap: str = PyDARNColormaps.PYDARN_VELOCITY,
                        filter_settings: dict = {},
                        date_fmt: str = '%y/%m/%d\n %H:%M', **kwargs):
        """
        Plots a range-time parameter plot of the given
        field name in the dmap_data

        Future Work
        -----------
        Support for other data input, like "time"
        dictionary key containing the datetime. However,
        further discussion is needed if this will be the keys
        name or maybe another input.

        Parameters
        -----------
        dmap_data: List[dict]
        parameter: str
            key name indicating which parameter to plot.
            Default: v (Velocity)
        beam_num : int
            The beam number of data to plot
            Default: 0
        channel : int or str
            The channel 0, 1, 2, 'all'
            Default : 'all'
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        groundscatter : boolean or str
            Flag to indicate if groundscatter should be plotted. If string
            groundscatter will be represented by that color else grey.
            Default : False
        background : str
            color of the background in the plot
            default: white
        zmin: int
            Minimum normalized value
            Default: minimum parameter value in the data set
        zmax: int
            Maximum normalized value
            Default: maximum parameter value in the data set
        ymax: int
            Sets the maximum y value
            Default: None, uses 'nrang' from data
        norm: matplotlib.colors.Normalization object
            This object use dependency injection to use any normalization
            method with the zmin and zmax.
            Default: colors.Normalization()
        start_time: datetime
            Start time of the plot x-axis as a datetime object
            Default: rounded to nearest hour-30 minutes from the first
                     record containing the chosen parameters data
        end_time: datetime
            End time of the plot x-axis as a datetime object
            Default: last record of the chosen parameters data
        date_fmt : str
            format of x-axis date ticks, follow datetime format
            Default: '%y/%m/%d\n %H:%M' (Year/Month/Day Hour:Minute)
        colorbar: matplotlib.pyplot.colorbar
            Setting a predefined colorbar for the range-time plot
            If None, then a colorbar will be created for the plot
            Default: None
        colorbar_label: str
            the label that appears next to the color bar
            Default: ''
        cmap: str or matplotlib.cm
            matplotlib colour map
            https://matplotlib.org/tutorials/colors/colormaps.html
            Default: PyDARNColormaps.PYDARN_VELOCITY
            note: to reverse the color just add _r to the string name
        plot_filter: dict
            dictionary of the following keys for filtering data out:
            max_array_filter : dict
                dictionary that contains the key parameter names and the values
                to compare against. Will filter out any data points
                that is above this value.
            min_array_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out any data points that is
                below this value.
            max_scalar_filter : dict
                dictionary that contains the key parameter names and the values
                to compare against. Will filter out data sections that is
                above this value.
            min_scalar_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out data sections
                that is below this value.
            equal_scalar_filter : dict
                dictionary that contains the key parameter names and the value
                to compare against. Will filter out data sections
                that is does not equal the value.
        kwargs:
            key names of variable settings of pcolormesh:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.pcolormesh.html

        Raises
        ------
        RTPUnknownParameterError
        RTPIncorrectPlotMethodError
        RTPNoDataFoundError
        IndexError

        Returns
        -------
        im: matplotlib.pyplot.pcolormesh
            matplotlib object from pcolormesh
        cb: matplotlib.colorbar
            matplotlib color bar
        cmap: matplotlib.cm
            matplotlib color map object
        time_axis: list
            list representing the x-axis datetime objects
        y_axis: list
            list representing the y-axis range gates
        z_data: 2D numpy array
            2D array of the parameters values at the given time and range gate

        See Also
        ---------
        colors: https://matplotlib.org/2.0.2/api/colors_api.html
        color maps: PyDARNColormaps or
                    https://matplotlib.org/tutorials/colors/colormaps.html
        normalize:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.colors.Normalize.html
        colorbar:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.colorbar.html
        pcolormesh:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.pcolormesh.html

        """
        # Settings
        plot_filter = {'min_array_filter': dict(),
                       'max_array_filter': dict(),
                       'min_scalar_filter': dict(),
                       'max_scalar_filter': dict(),
                       'equal_scalar_filter': dict()}

        plot_filter.update(filter_settings)

        # If an axes object is not passed in then store
        # the equivalent object in matplotlib. This allows
        # for variant matplotlib plotting styles.
        if not ax:
            ax = plt.gca()

        # Determine if a DmapRecord was passed in, instead of a list
        try:
            # because of partial records we need to find the first
            # record that has that parameter
            index_first_match = next(i for i, d in enumerate(dmap_data)
                                     if parameter in d)
            if isinstance(dmap_data[index_first_match][parameter],
                          DmapArray) or\
               isinstance(dmap_data[index_first_match][parameter],
                          DmapScalar):
                dmap_data = dmap2dict(dmap_data)
        except StopIteration:
            raise rtp_exceptions.RTPUnknownParameterError(parameter)
        cls.dmap_data = dmap_data
        cls.__check_data_type(parameter, 'array', index_first_match)
        start_time, end_time = cls.__determine_start_end_time(start_time,
                                                              end_time)

        # y-axis coordinates, i.e., range gates,
        # TODO: implement variant other coordinate systems for the y-axis
        # y shape needs to be +1 longer as requirement of how pcolormesh
        # draws the pixels on the grid

        # because nrang can change based on mode we need to look
        # for the largest value
        y_max = max(record['nrang'] for record in cls.dmap_data)
        y = np.arange(0, y_max+1, 1)

        # z: parameter data mapped into the color mesh
        z = np.zeros((1, y_max)) * np.nan

        # x: time date data
        x = []

        # We cannot simply use numpy's built in min and max function
        # because of the groundscatter value :(

        # These flags indicate if zmin and zmax should change
        set_zmin = True
        set_zmax = True
        if zmin is None:
            zmin = cls.dmap_data[index_first_match][parameter][0]
            set_zmin = False
        if zmax is None:
            zmax = cls.dmap_data[index_first_match][parameter][0]
            set_zmax = False

        for dmap_record in cls.dmap_data:
            # get time difference to test if there is some gap data
            rec_time = cls.__time2datetime(dmap_record)
            diff_time = 0.0
            if rec_time > end_time:
                break
            if x != []:
                # 60.0 seconds in a minute
                delta_diff_time = (rec_time - x[-1])
                diff_time = delta_diff_time.seconds/60.0

            # separation roughly 2 minutes
            if diff_time > 2.0:
                # if there is gap data (no data recorded past 2 minutes)
                # then fill it in with white space
                for _ in range(0, int(np.floor(diff_time/2.0))):
                    x.append(x[-1] + timedelta(0, 120))
                    i = len(x) - 1  # offset since we start at 0 not 1
                    if i > 0:
                        z = np.insert(z, len(z), np.zeros(1, y_max) * np.nan,
                                      axis=0)
            # Get data for the provided beam number
            if (beam_num == 'all' or dmap_record['bmnum'] == beam_num) and\
               (channel == 'all' or
                    dmap_record['channel'] == channel):
                if start_time <= rec_time:
                    # construct the x-axis array
                    # Numpy datetime is used because it properly formats on the
                    # x-axis
                    x.append(rec_time)
                    # I do this to avoid having an extra loop to just count how
                    # many records contain the beam number
                    i = len(x) - 1  # offset since we start at 0 not 1

                    # insert a new column into the z_data
                    if i > 0:
                        z = np.insert(z, len(z), np.zeros(1, y_max) * np.nan,
                                      axis=0)
                    try:
                        if len(dmap_record[parameter]) == dmap_record['nrang']:
                            good_gates = range(len(dmap_record[parameter]))
                        else:
                            good_gates = dmap_record['slist']

                        # get the range gates that have "good" data in it
                        for j in range(len(good_gates)):
                            # if it is groundscatter store a very
                            # low number in that cell
                            if groundscatter and\
                               dmap_record['gflg'][j] == 1:
                                # chosen value from davitpy to make the
                                # groundscatter a different color
                                # from the color map
                                z[i][good_gates[j]] = -1000000
                            # otherwise store parameter value
                            # TODO: refactor and clean up this code
                            elif cls.__filter_data_check(dmap_record,
                                                         plot_filter, j):
                                z[i][good_gates[j]] = \
                                        dmap_record[parameter][j]
                                # calculate min and max value
                                if not set_zmin and\
                                   z[i][good_gates[j]] < zmin:
                                    zmin = z[i][good_gates[j]]
                                if not set_zmax and \
                                   z[i][good_gates[j]] > zmax:
                                    zmax = z[i][good_gates[j]]
                    # a KeyError may be thrown because slist is not created
                    # due to bad quality data.
                    except KeyError:
                        continue
        x.append(end_time)
        # Check if there is any data to plot
        if np.all(np.isnan(z)):
            raise rtp_exceptions.RTPNoDataFoundError(parameter, beam_num,
                                                     start_time, end_time,
                                                     cls.dmap_data[0]['bmnum'])
        time_axis, y_axis = np.meshgrid(x, y)
        z_data = np.ma.masked_where(np.isnan(z.T), z.T)
        norm = norm(zmin, zmax)
        if isinstance(cmap, str):
            cmap = cm.get_cmap(cmap)

        if isinstance(groundscatter, str):
            cmap.set_under(groundscatter, 1.0)
        elif groundscatter:
            cmap.set_under('grey', 1.0)

        # set the background color, this needs to happen to avoid
        # the overlapping problem that occurs
        # cmap.set_bad(color=background, alpha=1.)
        # plot!
        im = ax.pcolormesh(time_axis, y_axis, z_data, lw=0.01,
                           cmap=cmap, norm=norm, **kwargs)
        # setup some standard axis information
        # Upon request of Daniel Billet and others, I am rounding
        # the time down so the plotting x-axis will show the origin
        # time label
        # TODO: may need to be its own function
        rounded_down_start_time = x[0] -\
            timedelta(minutes=x[0].minute % 15,
                      seconds=x[0].second,
                      microseconds=x[0].microsecond)
        ax.set_xlim([rounded_down_start_time, x[-1]])
        ax.xaxis.set_major_formatter(dates.DateFormatter(date_fmt))
        if ymax is None:
            ymax = y_max
        ax.set_ylim(0, ymax)
        ax.yaxis.set_ticks(np.arange(0, ymax+1, (ymax)/5))

        # SuperDARN file typically are in 2hr or 24 hr files
        # to make the minute ticks sensible, the time length is detected
        # then a interval is picked. 30 minute ticks for 24 hr plots
        # and 5 minute ticks for 2 hour plots.
        data_time_length = end_time - start_time
        # 3 hours * 60 minutes * 60 seconds
        if data_time_length.total_seconds() > 3*60*60:
            tick_interval = 30
        else:
            tick_interval = 1
        ax.xaxis.set_minor_locator(dates.MinuteLocator(interval=tick_interval))
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(5))
        # so the plots gets to the ends
        ax.margins(0)

        # create color bar if True
        if not colorbar:
            with warnings.catch_warnings():
                warnings.filterwarnings('error')
                try:
                    locator = ticker.MaxNLocator(symmetric=True, min_n_ticks=3,
                                                 integer=True, nbins='auto')
                    ticks = locator.tick_values(vmin=zmin, vmax=zmax)
                    cb = ax.figure.colorbar(im, ax=ax, extend='both',
                                            ticks=ticks)

                except (ZeroDivisionError, Warning):
                    raise rtp_exceptions.RTPZeroError(parameter, beam_num,
                                                      zmin, zmax,
                                                      norm) from None
        if colorbar_label != '':
            cb.set_label(colorbar_label)

        return im, cb, cmap, x, y, z_data

    @classmethod
    def plot_time_series(cls, dmap_data: List[dict],
                         parameter: str = 'tfreq', beam_num: int = 0,
                         ax=None, start_time: datetime = None,
                         end_time: datetime = None,
                         date_fmt: str = '%y/%m/%d\n %H:%M',
                         channel='all', scale: str = 'linear',
                         cp_name: bool = True, **kwargs):
        """
        Plots the time series of a scalar parameter

        Parameters
        ----------
        dmap_data : List[dict]
            List of dictionaries representing SuperDARN data
        parameter : str
            Scalar parameter to plot
            Default: tfreq
        beam_num : int
            The beam number of data to plot
            Default: 0
        ax : matplotlib axes object
            option to pass in axes object from matplotlib.pyplot
            Default: plt.gca()
        start_time: datetime
            Start time of the plot x-axis as a datetime object
            Default: first record date
        end_time: datetime
            End time of the plot x-axis as a datetime object
            Default: last record date
        date_fmt : datetime format string
            Date format for the x-axis
            Default: '%y/%m/%d \n %H:%M'
        channel : int or str
            integer indicating which channel to plot or 'all' to
            plot all channels
            Default: 'all'
        scale: str
            The y-axis scaling. This is not used for plotting the cp ID
            Default: log
        cp_name : bool
            If True, the cp ID name will be printed
            along side the number. Otherwise the cp ID will
            just be printed. This is only used for the parameter cp
            Default: True
        kwargs
            kwargs passed into plot_date

        Raises
        ------
        RTPUnknownParameterError
        RTPIncorrectPlotMethodError
        RTPNoDataFoundError
        IndexError

        Returns
        -------
        lines: list
            list of matplotlib.lines.Lines2D object representing the
            time-series data if plotting parameter cp then it will be None
        x: list
            list of datetime objects representing x-axis time series
        y: list
            list of scalar values for each datetime object

        See Also
        --------
        yscale:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.yscale.html
        plot_date:
            https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.axes.Axes.plot_date.html
        colors: https://matplotlib.org/2.0.2/api/colors_api.html
        """
        # check if axes object is passed in, if not
        # Default to plt.gca()
        if not ax:
            ax = plt.gca()

        # Determine if a DmapRecord was passed in, instead of a list
        try:
            # because of partial records we need to find the first
            # record that has that parameter
            index_first_match = next(i for i, d in enumerate(dmap_data)
                                     if parameter in d)
            if isinstance(dmap_data[index_first_match][parameter],
                          DmapArray) or\
               isinstance(dmap_data[index_first_match][parameter],
                          DmapScalar):
                dmap_data = dmap2dict(dmap_data)
        except StopIteration:
            raise rtp_exceptions.RTPUnknownParameterError(parameter)

        cls.dmap_data = dmap_data
        cls.__check_data_type(parameter, 'scalar', index_first_match)
        start_time, end_time = cls.__determine_start_end_time(start_time,
                                                              end_time)

        # initialized here for return purposes
        lines = None
        # parameter data
        y = []
        # date time
        x = []
        # plot CPID
        if parameter == 'cp':
            old_cpid = None
            for dmap_record in cls.dmap_data:
                # TODO: this check could be a function call
                x.append(cls.__time2datetime(dmap_record))

                if (dmap_record['bmnum'] == beam_num or beam_num == 'all') and\
                   (dmap_record['channel'] == channel or channel == 'all'):
                    rec_time = cls.__time2datetime(dmap_record)
                    if start_time <= rec_time and rec_time <= end_time:
                        if old_cpid != dmap_record['cp'] or old_cpid is None:
                            ax.axvline(x=rec_time, color='black')
                            old_cpid = dmap_record['cp']
                            ax.text(x=rec_time + timedelta(seconds=600), y=0.7,
                                    s=dmap_record['cp'])
                            if cp_name:
                                # Keeping this commented code in to show how
                                # we could get the name from the file; however,
                                # there is not set format for combf field ...
                                # so we will use the dictionary to prevent
                                # errors or incorrect names on the plot.
                                # However, we should get it from the file
                                # not a dictionary that might not be updated
                                # cpid_command =
                                #   dmap_record['combf'].split(' ')
                                # if len(cpid_command) == 1:
                                #     cp_name = cpid_command[0]
                                # elif len(cpid_command) == 0:
                                #     cp_name = 'unknown'
                                # else:
                                #     cp_name = cpid_command[1]
                                if dmap_record['cp'] < 0:
                                    cpID_name = 'discretionary \n{}'\
                                            ''.format(SuperDARNCpids.cpids.
                                                      get(abs(dmap_record['cp']),
                                                          'unknown'))
                                else:
                                    cpID_name =\
                                            SuperDARNCpids.cpids.\
                                            get(abs(dmap_record['cp']),
                                                'unknown')
                                ax.text(x=rec_time + timedelta(seconds=600),
                                        y=0.1, s=cpID_name)

            # Check if the old cp ID change, if not then there was no data
            if old_cpid is None:
                raise rtp_exceptions.\
                        RTPNoDataFoundError(parameter, beam_num,
                                            start_time, end_time,
                                            cls.dmap_data[0]['bmnum'])

            # to get rid of y-axis numbers
            ax.set_yticks([])
        else:
            for dmap_record in cls.dmap_data:
                # TODO: this check could be a function call
                rec_time = cls.__time2datetime(dmap_record)
                if start_time <= rec_time and rec_time <= end_time:
                    if (dmap_record['bmnum'] == beam_num or
                        beam_num == 'all') and \
                       (channel == dmap_record['channel'] or channel == 'all'):
                        # construct the x-axis array
                        x.append(rec_time)
                        if parameter == 'tfreq':
                            # Convert kHz to MHz by dividing by 1000
                            y.append(dmap_record[parameter]/1000)
                        else:
                            y.append(dmap_record[parameter])
                    # else plot missing data
                    elif len(x) > 0:
                        diff_time = rec_time - x[-1]
                        # if the time difference is greater than 2 minutes
                        # meaning no data was collected for that time period
                        # then plot nothing.
                        if diff_time.total_seconds() > 2.0 * 60.0:
                            x.append(rec_time)
                            y.append(np.nan)  # for masking the data
            # Check if there is any data to plot
            if np.all(np.isnan(y)) or len(x) == 0:
                raise rtp_exceptions.\
                        RTPNoDataFoundError(parameter, beam_num,
                                            start_time, end_time,
                                            cls.dmap_data[0]['bmnum'])

            # using masked arrays to create gaps in the plot
            # otherwise the lines will connect in gapped data
            my = np.ma.array(y)
            my = np.ma.masked_where(np.isnan(my), my)
            lines = ax.plot_date(x, my, fmt='k', tz=None, xdate=True,
                                 ydate=False,
                                 **kwargs)
            rounded_down_start_time = x[0] -\
                timedelta(minutes=x[0].minute % 15,
                          seconds=x[0].second,
                          microseconds=x[0].microsecond)
            ax.set_xlim([rounded_down_start_time, x[-1]])
            ax.set_yscale(scale)

        # set date format and minor hourly locators
        # Rounded the time down to show origin label upon
        # Daniel Billet and others request.
        # TODO: may move this to its own function
        rounded_down_start_time = x[0] -\
            timedelta(minutes=x[0].minute % 15,
                      seconds=x[0].second,
                      microseconds=x[0].microsecond)
        ax.set_xlim([rounded_down_start_time, x[-1]])

        ax.xaxis.set_major_formatter(dates.DateFormatter(date_fmt))
        ax.xaxis.set_minor_locator(dates.HourLocator())
        # SuperDARN file typically are in 2hr or 24 hr files
        # to make the minute ticks sensible, the time length is detected
        # then a interval is picked. 30 minute ticks for 24 hr plots
        # and 5 minute ticks for 2 hour plots.
        data_time_length = end_time - start_time
        # 3 hours * 60 minutes * 60 seconds
        if data_time_length.total_seconds() > 3*60*60:
            tick_interval = 30
        else:
            tick_interval = 1
        ax.xaxis.set_minor_locator(dates.MinuteLocator(interval=tick_interval))

        ax.margins(x=0)
        ax.tick_params(axis='y', which='minor')
        return lines, x, y

    @classmethod
    def plot_summary(cls, dmap_data: List[dict], beam_num: int = 0,
                     groundscatter: bool = True, channel: int = 'all',
                     figsize: tuple = (11, 8.5), watermark: bool = True,
                     boundary: dict = {}, background_color: str = 'w',
                     cmaps: dict = {}, lines: dict = {},
                     plot_elv: bool = True, title=None):
        """
        Plots the summary of several SuperDARN parameters using time-series and
        range-time plots. Please see Notes for further description
        on what is plotted.

        Future Work
        ------------
        day-night terminators
        slant ranges

        Parameters
        ----------
        dmap_data: List[dict]
            List of dictionaries of the data to be plotted containing the
            parameter fields used in the summary plot.
        beam_num : int
            beam number to plot
            default: 0
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        groundscatter : boolean
            Flag to indicate if groundscatter should be plotted.
            Placed only on the velocity plot.
            Default : True
        channel : int
            channel number that will be plotted
            in the summary plot.
            Default: 'all'
        figsize : (int,int)
            tuple containing (height, width) figure size
            Default: 11 x 8.5
        watermark : boolean
            text that runs across the plot stating "Not for Publication Use"
            default: True
        cmaps: dict or str
            dictionary of matplotlib color maps for the summary
            range time parameter plots.
            https://matplotlib.org/tutorials/colors/colormaps.html
            Default: {'p_l': 'plasma',
                      'v': PyDARNColormaps.PYDARN_VELOCITY,
                      'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                      'elv': PyDARNColormaps.PYDARN}
            note: to reverse the color just add _r to the string name
        lines: dict or str
            dictionary of time-series line colors.
            Default: black
        background_color: str
            changes the color of the background in the plots.
            Default: white
        boundary: dict
            tuple as the value (zmin, zmax) for the plots
            Default: {'noise.sky': (1e0, 1e5),
                      'tfreq': (8, 22),
                      'nave': (0, 60),
                      'p_l': (0, 45),
                      'v': (-200, 200),
                      'w_l': (0, 250),
                      'elv': (0, 45)}
        plot_elv: boolean
            boolean determines if elevation should be plotted or not.
            If there is no data for elevation data field then elevation is not
            plotted.
            Default: True
        title: str
            title of the plot
            Default: auto-generated by the files details
            {radar name} {Radar system (if applicable)} Fitacf {version}
            {start hour/date} - {end hour/date}  Beam {number}

        Raises
        ------
        IndexError
        RTPUnknownParameterError
        RTPIncorrectPlotMethodError
        RTPNoDataFoundError

        See Also
        --------
        plot_range_time : plots range-time parameter
        plot_time_series : plots time-series

        Return
        ------
        fig: matplotlib.pyplot.figure object
        axes: list
            list of matplotlib.pyplot.axes objects to generate
            the subplots in the summary plot

        Notes
        -----
        These are the following parameters in the SuperDARN FitACF file that is
        plotted in a summary plot:
            - noise.sky :  (time-series)
            - tfreq : transmission frequency (time-series)
            - nave : number of averages  (time-series)
            - cp : control program ID (time-series)
            - p_l : Signal to Noise ratio (range-time)
            - v : velocity (range-time)
            - w_l : spectral width (range-time)
            - elv : elevation (optional) (range-time)
        """
        message = "WARNING: matplotlib Default dpi may cause distortion"\
                  " in range gates and time period. The figure size can"\
                  " be adjusted with the option figsize and dpi can be"\
                  " adjusted when saving the file."
        warnings.warn(message)

        # Default boundary ranges for the various parameter
        boundary_ranges = {'noise.search': (1e0, 1e5),
                           'noise.sky': (1e0, 1e5),
                           'tfreq': (8, 22),
                           'nave': (0, 60),
                           'p_l': (0, 45),
                           'v': (-200, 200),
                           'w_l': (0, 250),
                           'elv': (0, 45)}
        boundary_ranges.update(boundary)

        # Default color maps for the summary plot
        line = {'noise.search': 'k',
                'noise.sky': 'k',
                'tfreq': 'k',
                'nave': 'k'}

        if isinstance(lines, dict):
            line.update(lines)
        else:
            line.update({k: lines for k, v in line.items()})
        cmap = {'p_l': 'plasma',
                'v': PyDARNColormaps.PYDARN_VELOCITY,
                'w_l': PyDARNColormaps.PYDARN_VIRIDIS,
                'elv': PyDARNColormaps.PYDARN}
        if isinstance(cmaps, dict):
            cmap.update(cmaps)
        else:
            cmap.update({k: cmaps for k, v in cmap.items()})

        fig = plt.figure(figsize=figsize)

        # axes objects in order of creation:
        # [noise, tfreq, cp, snr, vel, spect, elv]
        # Check if the radar has elevation information if not
        # do not plot elevation
        try:
            # need to use any because some records at the start
            # can be partial which doesn't mean there is no elv
            # data
            if any('elv' in d for d in dmap_data) and plot_elv:
                num_plots = 7
            else:
                num_plots = 6
        except KeyError:
            num_plots = 6
        axes = []
        # List of parameters plotted in the summary plot, tuples are used
        # for shared plot parameters like noise.search and noise.sky
        # Removing search noise from the summary plot due to too much
        # overlapping, however, if confirmed as undesired parameter to
        # plot, will remove it from the code.
        # TODO: remove search noise completely if not wanted in summary plot
        axes_parameters = [('noise.sky', 'noise.search'), ('tfreq', 'nave'),
                           ('cp'), ('p_l'), ('v'), ('w_l'), ('elv')]
        # labels to show on the summary plot for each parameter
        labels = [('Sky \n Noise', 'Search\n Noise'),
                  ('Freq\n ($MHz$)', 'Nave'), ('CP ID'), ('SNR ($dB$)'),
                  ('Velocity\n ($m\ s^{-1}$)'),
                  ('Spectral Width\n ($m\ s^{-1}$)'),
                  ('Elevation\n ($\degree$)')]

        for i in range(num_plots):
            # time-series plots
            # position: [left, bottom, width, height]
            if i < 3:
                axes.append(fig.add_axes([0.1, 0.88 - (i*0.08), 0.76, 0.06]))
            # range-time plots
            else:
                axes.append(fig.add_axes([0.1, 1.04 - (i*0.16), 0.95, 0.14]))

        for i in range(num_plots):
            # plot time-series
            if i < 2:
                # for noise.search and frequency plots as they share x-axis
                # with noise.sky and nave
                if i == 0:
                    scale = 'log'
                else:
                    scale = 'linear'

                # plot time-series parameters that share a plot
                if i < 2:
                    with warnings.catch_warnings(record=True) as w:
                        cls.plot_time_series(dmap_data, beam_num=beam_num,
                                             parameter=axes_parameters[i][0],
                                             channel=channel, scale=scale,
                                             color=line[axes_parameters[i][0]],
                                             ax=axes[i], linestyle='-',
                                             label=labels[i][0])
                    if len(w) > 0:
                        warnings.\
                            warn("Warning: {parameter} raised the"
                                 " following warning: {message}"
                                 "".format(parameter=axes_parameters[i][0],
                                           message=str(w[0].message)))
                    axes[i].set_ylabel(labels[i][0], rotation=0, labelpad=30)
                    axes[i].\
                        axhline(y=boundary_ranges[axes_parameters[i][0]][0] +
                                0.8, xmin=-0.11, xmax=-0.05, clip_on=False,
                                color=line[axes_parameters[i][0]])
                    axes[i].set_ylim(boundary_ranges[axes_parameters[i][0]][0],
                                     boundary_ranges[axes_parameters[i][0]][1])
                    # For better y-axis ticks
                    if scale == 'log':
                        axes[i].yaxis.set_major_locator(ticker.
                                                        LogLocator(numticks=3))
                    else:
                        axes[i].yaxis.\
                                set_major_locator(ticker.
                                                  MaxNLocator(integer=True,
                                                              nbins=3))
                    axes[i].yaxis.set_label_coords(-0.08, 0.085)

                    if i == 1:
                        # plot the shared parameter
                        second_ax = axes[i].twinx()
                        # warnings are not caught with try/except
                        with warnings.catch_warnings(record=True) as w:
                            cls.plot_time_series(dmap_data, beam_num=beam_num,
                                                 parameter=axes_parameters[i][1],
                                                 color=line[axes_parameters[i][1]],
                                                 channel=channel,
                                                 scale=scale, ax=second_ax,
                                                 linestyle='--')
                        if len(w) > 0:
                            warnings.\
                                warn("Warning: {parameter} raised the"
                                     " following warning: {message}"
                                     "".format(parameter=axes_parameters[i][1],
                                               message=str(w[0].message)))
                        second_ax.set_xticklabels([])
                        second_ax.set_ylabel(labels[i][1], rotation=0,
                                             labelpad=25)
                        second_ax.\
                            axhline(y=boundary_ranges[axes_parameters[i][1]][0]
                                    + 0.8, xmin=1.07, xmax=1.13,
                                    clip_on=False, linestyle='--',
                                    color=line[axes_parameters[i][1]])
                        second_ax.\
                            set_ylim(boundary_ranges[axes_parameters[i][1]][0],
                                     boundary_ranges[axes_parameters[i][1]][1])
                        second_ax.yaxis.set_label_coords(1.1, 0.7)
                        if scale == 'log':
                            second_ax.yaxis.\
                                    set_major_locator(ticker.
                                                      LogLocator(numticks=4))

                        else:
                            second_ax.yaxis.\
                                    set_major_locator(ticker.
                                                      MaxNLocator(integer=True,
                                                                  nbins=3))

                axes[i].set_facecolor(background_color)
            # plot cp id
            elif i == 2:
                cls.plot_time_series(dmap_data, beam_num=beam_num,
                                     parameter=axes_parameters[i],
                                     channel=channel,
                                     ax=axes[i])
                axes[i].set_ylabel('CPID', rotation=0, labelpad=30)
                axes[i].yaxis.set_label_coords(-0.08, 0.079)
                axes[i].set_facecolor(background_color)
            # plot range-time
            else:
                # Current standard is to only have groundscatter
                # on the velocity plot. This may change in the future.
                if groundscatter and axes_parameters[i] == 'v':
                    grndflg = True
                else:
                    grndflg = False
                _, cbar, _, x, _, _ =\
                    cls.plot_range_time(dmap_data,
                                        beam_num=beam_num,
                                        colorbar_label=labels[i],
                                        parameter=axes_parameters[i],
                                        ax=axes[i],
                                        groundscatter=grndflg,
                                        channel=channel,
                                        cmap=cmap[axes_parameters[i]],
                                        zmin=boundary_ranges[axes_parameters[i]][0],
                                        zmax=boundary_ranges[axes_parameters[i]][1],
                                        ymax=75,
                                        background=background_color)
                # Overwriting velocity ticks to get a better pleasing
                # look on the colorbar
                # Preference by Marina Schmidt
                if axes_parameters[i] == 'v':
                    locator = ticker.LinearLocator(numticks=5)
                    ticks =\
                        locator.tick_values(vmin=boundary_ranges[axes_parameters[i]][0],
                                            vmax=boundary_ranges[axes_parameters[i]][1])
                    if ticks[0] < boundary_ranges[axes_parameters[i]][0]:
                        ticks[0] = boundary_ranges[axes_parameters[i]][0]

                    if ticks[-1] > boundary_ranges[axes_parameters[i]][1]:
                        ticks[-1] = boundary_ranges[axes_parameters[i]][1]
                    cbar.set_ticks(ticks)
                axes[i].set_ylabel('Range Gates')
            if i < num_plots-1:
                axes[i].set_xticklabels([])
            # last plot needs the label on the x-axis
            else:
                axes[i].set_xlabel('Date (UTC)')

        if title is None:
            plt.title(cls.__generate_title(x[0], x[-1], beam_num,
                                           channel), y=2.4)
        else:
            plt.title(title, y=2.4)
        plt.subplots_adjust(wspace=0, hspace=0)
        if watermark:
            fig.text(0.90, 0.99, "Not for Publication Use", fontsize=75,
                     color='gray', ha='right', va='top',
                     rotation=-38, alpha=0.3)

        return fig, axes

    @classmethod
    def __generate_title(cls, start_time: datetime, end_time: datetime,
                         beam_num: int, channel: int) -> str:
        if cls.dmap_data[0]['fitacf.revision.major'] == 5:
            version = "2.5"
        else:
            version = "{major}.{minor}"\
                    "".format(major=cls.dmap_data[0]['fitacf.revision.major'],
                              minor=cls.dmap_data[0]['fitacf.revision.minor'])
        if cls.dmap_data[0]['origin.code'] == 100:
            radar_system = " (Borealis)"
        else:
            # I would put ROS but Alaska and AGILE DARN might have their own
            # systems and not sure how to decipher between them. If something
            # changes in the file structure, then I can add it here.
            radar_system = ""
        radar_name = SuperDARNRadars.radars[cls.dmap_data[0]['stid']].name
        # Date time formats:
        #   %Y - year
        #   %b - month abbreviation
        #   %d - day
        #   %H - Hour
        #   %M - 2-digit minutes
        if end_time.day == start_time.day:
            end_format = "%H:%M"
        elif end_time.month == start_time.month:
            end_format = "%d %H:%M"
        elif end_time.year == start_time.year:
            end_format = "%b %d %H:%M"
        else:
            end_format = "%Y %b %d %H:%M"
        title_format =\
            "{name}{system} Fitacf {version}"\
            "{system} {start_date} - {end_date}  Beam {num}"\
            "".format(name=radar_name, version=version,
                      system=radar_system,
                      start_date=start_time.strftime("%Y %b %d %H:%M"),
                      end_date=end_time.strftime(end_format),
                      num=beam_num)
        if type(channel) is int:
            title_format += " channel {ch_num}".format(ch_num=channel)
        return title_format

    @classmethod
    # TODO: could parallelize this method
    def __filter_data_check(cls, dmap_record: dict,
                            settings: dict, j: int) -> bool:
        """
        checks for data that does not meet the criteria of the filtered
        settings

        Parameters
        ----------
        dmap_record : dict
            dictionary of the dmap record fields
        settings : dict
            dictionary of the settings list
        j : int
            index on the slist or range gate to check the array values

        Returns
        -------
        pass_flg : bool
            boolean indicating if the dmap_record passes all filter checks
        """
        pass_flg = True
        for key, value in settings['min_array_filter'].items():
            if dmap_record[key][j] < value:
                pass_flg = False
                break
        for key, value in settings['max_array_filter'].items():
            if dmap_record[key][j] > value:
                pass_flg = False
                break

        for key, value in settings['min_scalar_filter'].items():
            if dmap_record[key] < value:
                pass_flg = False
                break

        for key, value in settings['max_scalar_filter'].items():
            if dmap_record[key] > value:
                pass_flg = False
                break

        for key, value in settings['equal_scalar_filter'].items():
            if dmap_record[key] != value:
                pass_flg = False
                break

        return pass_flg

    @classmethod
    def __check_data_type(cls, parameter: str, expected_type: str, i: int):
        """
        Checks to make sure the plot type is correct
        for the data structure

        Parameters
        ----------
        parameter: str
            string key word name of the parameter
        expected_type: str
            string describing an array or scalar type
            to determine which one is needed for the type of plot

        Raises
        -------
        RTPIncorrectPlotMethodError
        """
        data_type = cls.dmap_data[i][parameter]
        if expected_type == 'array':
            if not isinstance(data_type, np.ndarray):
                raise rtp_exceptions.RTPIncorrectPlotMethodError(parameter,
                                                                 data_type)
        else:
            if isinstance(data_type, np.ndarray):
                raise rtp_exceptions.RTPIncorrectPlotMethodError(parameter,
                                                                 data_type)

    # TODO: move to a utils or SuperDARN utils
    @classmethod
    def __time2datetime(cls, dmap_record: dict) -> datetime:
        """
        Converts DMAP time parameter fields into a datetime object

        Parameter
        ---------
        dmap_record: dict
            dictionary of the DMAP data contains the time data

        Returns
        -------
        datetime object
            returns a datetime object of the records time stamp
        """
        year = dmap_record['time.yr']
        month = dmap_record['time.mo']
        day = dmap_record['time.dy']
        hour = dmap_record['time.hr']
        minute = dmap_record['time.mt']
        second = dmap_record['time.sc']
        micro_sec = dmap_record['time.us']

        return datetime(year=year, month=month, day=day, hour=hour,
                        minute=minute, second=second, microsecond=micro_sec)

    # TODO: if used in other plotting methods then this should moved to
    #       utils
    @classmethod
    def __determine_start_end_time(cls, start_time: datetime,
                                   end_time: datetime) -> tuple:
        """
        Sets the start and end time based on import of dmap_data

        Parameter
        ---------
        start_time: datetime
            Start time is used to check if it was set or not
        end_time: datetime
            End time is used to check if it was set or not

        Returns
        -------
        start_time: datetime
        end_time: datetime
        """
        if not start_time:
            start_time = cls.__time2datetime(cls.dmap_data[0])
        if not end_time:
            end_time = cls.__time2datetime(cls.dmap_data[-1])
        return start_time, end_time



# copyright (C) SuperDARN Canada, University of Saskatchewan
# Author(s): Marina Schmidt
#
# Modifications:
#
# Disclaimer: pyDARN is licensed under the GPL v 3.0 found in LICENSE
#


def standard_warning_format(message: str, category: str, filename: str,
                            lineno: int, file: str = None,
                            line: int = None) -> str:
    """
    Sets the standard warning message format to be:
        filename: lineno: category: message

    Parameters
    ----------
        message: str
            warning message to be printed to a user
        category: str
            type of the warning message
        filename: str
            name of the associated to the warning message
        lineno : int
            line number where the warning message was raised
        file : str
        line : str

    Returns
    -------
        formatted warning message to be printed to the console
    """
    return "{filename}: {linenum}: {category}:"\
           " {message}\n".format(filename=filename,
                                 linenum=lineno,
                                 category=category.__name__,
                                 message=message)


def only_message_warning_format(message: str, category: str, filename: str,
                                lineno: int, file: str = None,
                                line: int = None) -> str:
    """
    Sets the standard warning message format to be:
        filename: lineno: category: message

    Parameters
    ----------
        message: str
            warning message to be printed to a user
        category: str
            type of the warning message
        filename: str
            name of the associated to the warning message
        lineno : int
            line number where the warning message was raised
        file : str
        line : str

    Returns
    -------
        formatted warning message to be printed to the console
    """
    return "{category}: {message}\n".format(filename=filename,
                                            linenum=lineno,
                                            category=category.__name__,
                                            message=message)



"""
Copyright 2018 SuperDARN Canada, Univeristy Saskatchewan
Author(s): Marina Schmidt

Licensed under GNU v3.0

__init__.py
2018-11-05
Init file to setup the logging configuration and linking pyDARN's
module, classes, and functions.
"""
# KEEP THIS FILE AS MINIMAL AS POSSIBLE!

import os
import logging.config
import yaml

# Importing pydarn exception classes
from .exceptions import dmap_exceptions
from .exceptions import superdarn_exceptions
from .exceptions import rtp_exceptions
from .exceptions import borealis_exceptions
from .exceptions import radar_exceptions
from .exceptions.warning_formatting import standard_warning_format
from .exceptions.warning_formatting import only_message_warning_format

# Importing pydarn pydmap data structure classes
from .io.datastructures import DmapScalar
from .io.datastructures import DmapArray
from .io import superdarn_formats

# importing utils
from .utils.conversions import dict2dmap
from .utils.conversions import dmap2dict
from .utils.superdarn_radars import SuperDARNRadars
from .utils.superdarn_cpid import SuperDARNCpids
from .utils.superdarn_radars import Hemisphere
from .utils.superdarn_radars import read_hdw_file

# Importing pydarn dmap classes
from .io.dmap import DmapRead
from .io.dmap import DmapWrite

# Importing pydarn superdarn classes
from .io.superdarn import SDarnRead
from .io.superdarn import SDarnWrite
from .io.superdarn import SDarnUtilities

# Importing pydarn borealis classes
from .io.borealis import borealis_formats
from .io.borealis.borealis import BorealisRead
from .io.borealis.borealis import BorealisWrite
from .io.borealis.borealis_convert import BorealisConvert

# import plotting
from .plotting.color_maps import PyDARNColormaps
from .plotting.rtp import RTP

"""
pyDARN uses yaml for logging configuration because its legibility
makes it the preferred configuration file format.
"""
# real path is needed because path imports from where it is ran and the
# logging config will not be in the users current path.
real_path = os.path.realpath(__file__)
dirname = os.path.dirname(real_path)

# setting the logging configuration.
log_path = dirname + "/logging_config.yaml"
with open(log_path, 'r') as f:
    config = yaml.safe_load(f.read())
    logging.config.dictConfig(config)



# Copyright (C) SuperDARN Canada, University of Saskatchewan
# Authors: Keith Kotyk and Marina Schmidt
import logging
pydarn_logger = logging.getLogger('pydarn')


class CursorError(Exception):
    """
    Raise if the cursor is not correctly set

    Parameter
    ---------
    cursor : int
        Current position in the data buffer
    expected_value : int
        What you expected the value of the cursor to be.
        Default 0.
    message : str
        Another type of message that might give more information
        on the cursor error.
        Default empty.
    """
    def __init__(self, cursor: int, expected_value: int = 0,
                 rec_num: int = 0, message=''):
        self.cursor = cursor
        if message == '':
            self.message = "Error: Cursor is at {cursor} and"\
                    "it needs to be {expected}. Failed at record {rec}"\
                    "".format(cursor=cursor, expected=expected_value,
                              rec=rec_num)
        else:
            self.message = message
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class EmptyFileError(Exception):
    """
    Raised if a file is empty.

    Parameter
    --------
    filename : str
        The name of the file that is empty.
    """
    def __init__(self, filename: str):
        self.filename = filename
        self.message = "Error: {} is empty or"\
            " please check this is the correct file".format(filename)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class DmapCharError(Exception):
    """
    Raised if a char type is str

    Parameter
    --------
    filename : str
        name of the file that the DMAP data is coming from.
    data_name : str
        parameter field name in the DMAP record
    rec_num : int
        Record number the error was raised at
    """
    def __init__(self, data_name: str, rec_num: int):
        self.data_name = data_name,
        self.rec_num = rec_num
        self.message = "Error: For field {field} at record number"\
            " {rec_num} is"\
            " a string type trying to be written in as a char."\
            " DMAP treats char as int8. Please revise this field"\
            " type".format(field=self.data_name,
                           rec_num=self.rec_num)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class DmapDataTypeError(Exception):
    """
    Raised if the data type is not correct.

    Parameter
    ---------
    filename : str
        name of the file that the DMAP data is coming from.
    data_name : str
        parameter field name in the DMAP record
    data_type : str
        data type of the field name.
    cursor : int
        current position in the data buffer
    """
    def __init__(self, filename: str, data_name: str,
                 data_type: str, rec_num: int):
        self.filename = filename
        self.message = "Error: Dmap data type {data_type} for {name}"\
            " at record {rec} does not exist in dmap data types."\
            " Filename: {filename}".format(name=data_name,
                                           data_type=data_type,
                                           filename=filename,
                                           rec=rec_num)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class NegativeByteError(Exception):
    """
    Raised if an element has < 0 bytes.

    Parameter
    ---------
    filename : str
        The file name of the DMAP data
    element_info : str
        String containing element name.
    """
    def __init__(self, filename: str, element_info: str, rec_num: int):
        self.filename = filename
        self.message = "Error: {filename} contains an {element} < 0"\
            " at record {rec}.".format(filename=filename,
                                       element=element_info,
                                       rec=rec_num)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class ZeroByteError(Exception):
    """
    Raised if an element has == 0 bytes.

    Parameter
    ---------
    filename : str
        Name of the file that contains the DMAP data
    element_info : str
        String containing element name.
    cursor : int
        Current position in the dmap buffer
    """
    def __init__(self, filename: str, element_info: str, rec_num: int):
        self.filename = filename
        self.message = "Error: {filename} contains an {element} == 0"\
            " at record {rec}.".format(filename=filename,
                                       element=element_info,
                                       rec=rec_num)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


class MismatchByteError(Exception):
    """
    Raised if an element is > another element.

    Parameter
    ---------
    filename : str
        Name of the file that contains the DMAP data
    element_info : str
        String containing element info on the two elements being compared.
    cursor : int
        Current position in the dmap buffer.
    """
    def __init__(self, filename: str, element_info: str, rec_num: int):
        self.filename = filename
        self.rec_num = rec_num
        self.message = "Error: {filename} contains an {element}"\
            " at record = {rec}.".format(filename=filename,
                                         element=element_info,
                                         rec=rec_num)
        super().__init__(self.message)
        pydarn_logger.error(self.message)


# TODO: may need to delete later if not used in DmapWrite?
class DmapDataError(Exception):
    """Raised if there is an error in parsing of data

    Parameter
    ---------
    filename : str
        Name of the file that contains the DMAP data
    message : str
        Message containing information on the Error
    """
    def __init__(self, filename: str, message: str):
        self.filename = filename
        self.message = " The following error for {filename} was raised: "\
            "{msg}".format(filename=self.filename,
                           msg=message)
        super().__init__(self.message)
        pydarn_logger.error("DmapDataError: {}".format(self.message))


class DmapTypeError(Exception):
    """
    Raised if there mismatch between DMAP data type

    Parameters
    ----------
    filename : str
        name of the file associated to the error
    data_type : str
        the data that does not match what DMAP can read
    rec_num : int
        the record number associated to the error
    """
    def __init__(self, filename: str, data_type: str, rec_num: int):
        self.filename = filename
        self.data_type = data_type
        self.rec_num = rec_num
        self.message = "Error: {data_type} does not match the DMAP data type"\
            " structures: DmapScalar, DmapArray."\
            " Please make sure you have the correct"\
            " Data Structure. Failed at record "\
            "{rec}".format(data_type=self.data_type,
                           filename=self.filename,
                           rec=self.rec_num)
        Exception.__init__(self, self.message)
        pydarn_logger.error("DataTypeError: {}".format(self.message))


class FilenameRequiredError(Exception):
    """
    Raised if a filename is not provided when needed for the procedure
    """
    def __init__(self):
        self.message = "Error: Filename is required"
        Exception.__init__(self, self.message)



import matplotlib.colors as mcol


class PyDARNColormaps():
    PYDARN_VELOCITY =\
            mcol.LinearSegmentedColormap.from_list("pydarn_velocity",
                                                   ["darkred", "r",
                                                    'pink', "b", "darkblue"])

    PYDARN_VIRIDIS =\
        mcol.LinearSegmentedColormap.from_list("pydarn_viridis",
                                               ["indigo", "midnightblue",
                                                "navy", "mediumblue",
                                                "teal",
                                                "mediumseagreen",
                                                "limegreen", "yellowgreen",
                                                "yellow", "gold"])

    PYDARN =\
        mcol.LinearSegmentedColormap.from_list("pydarn", ["midnightblue",
                                                          "darkblue",
                                                          "mediumblue",
                                                          "rebeccapurple",
                                                          "purple",
                                                          "darkmagenta",
                                                          "mediumvioletred",
                                                          "crimson", "red",
                                                          "orangered",
                                                          "darkorange",
                                                          "orange", "gold"])



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marci Detwiller
"""
This file contains classes for reading, writing, and
converting of Borealis site file types. Site file types
means Borealis site files, ie. stored in a record-by-record
fashion, before being converted to array types only.

Classes
-------
BorealisSiteRead: Reads Borealis site SuperDARN file types (hdf5)
BorealisSiteWrite: Writes Borealis SuperDARN file types (hdf5)

Exceptions
----------
BorealisFileTypeError
BorealisFieldMissingError
BorealisExtraFieldError
BorealisDataFormatTypeError

See Also
--------
BorealisUtilities
BorealisArrayRead
BorealisArrayWrite
BorealisRestructureUtilities

For more information on Borealis data files and how they convert to SDARN
files, see: https://borealis.readthedocs.io/en/latest/

Future Work
-----------
Add compression to bzip2

"""
import deepdish as dd
import h5py
import logging
import os
import subprocess as sp
import warnings

from collections import OrderedDict
from pathlib2 import Path
from typing import Union

from pydarn import borealis_exceptions, borealis_formats

from .borealis_utilities import BorealisUtilities
from .restructure_borealis import BorealisRestructureUtilities

pydarn_log = logging.getLogger('pydarn')


class BorealisSiteRead():
    """
    Class for reading Borealis 'site' filetypes. Site files are stored in
    a record-by-record style.

    See Also
    --------
    BorealisRawacf
    BorealisBfiq
    BorealisAntennasIq
    BorealisRawrf

    filename: str
        The filename of the Borealis HDF5 file being read.
    borealis_filetype: str
        The type of Borealis file. Types include:
        'bfiq'
        'antennas_iq'
        'rawacf'
        'rawrf'
    record_names: list(str)
    records: dict
    arrays: dict
    """

    def __init__(self, filename: str, borealis_filetype: str):
        """
        Reads Borealis site file types into a dictionary.

        Parameters
        ----------
        filename: str
            file name containing Borealis hdf5 data.
        borealis_filetype: str
            The type of Borealis file. Types include:
            'bfiq'
            'antennas_iq'
            'rawacf'
            'rawrf'

        Raises
        ------
        OSError
            Unable to open file
        """
        self.filename = filename
        if borealis_filetype not in ['bfiq', 'antennas_iq', 'rawacf', 'rawrf']:
            raise borealis_exceptions.BorealisFileTypeError(
                self.filename, borealis_filetype)
        self.borealis_filetype = borealis_filetype

        with h5py.File(self.filename, 'r') as f:
            self._record_names = sorted(list(f.keys()))
            # list of group names in the HDF5 file, to allow partial read.

        # Records are private to avoid tampering.
        self._records = OrderedDict()
        self.read_file()

    def __repr__(self):
        """ for representation of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class
        return "{class_name}({filename})"\
            "".format(class_name=self.__class__.__name__,
                      filename=self.filename)

    def __str__(self):
        """ for printing of the class object"""

        return "Reading from {filename} a total number of"\
            " records: {total_records}"\
            "".format(filename=self.filename,
                      total_records=len(list(self.records.keys())))

    @property
    def record_names(self):
        """
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms since
        epoch), and are equal to the group names in the site file types.
        """
        return self._record_names

    @property
    def records(self):
        """
        The Borealis data in a dictionary of records, according to the
        site file format.
        """
        return self._records

    @property
    def arrays(self):
        """
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
        """
        return BorealisRestructureUtilities.borealis_site_to_array_dict(
                                            self.filename, self.records,
                                            self.borealis_filetype)

    def read_file(self) -> dict:
        """
        Reads the specified Borealis file using the other functions for
        the proper file type.

        Reads the entire file.

        See Also
        --------
        read_bfiq
        read_rawacf
        read_antennas_iq
        read_rawrf

        Returns
        -------
        records: OrderedDict{dict}
            records of borealis data. Keys are first sequence timestamp
            (in ms since epoch).

        Raises
        ------
        BorealisFileTypeError
        """
        if self.borealis_filetype == 'bfiq':
            return self.read_bfiq()
        elif self.borealis_filetype == 'rawacf':
            return self.read_rawacf()
        elif self.borealis_filetype == 'antennas_iq':
            return self.read_antennas_iq()
        elif self.borealis_filetype == 'rawrf':
            return self.read_rawrf()
        else:
            raise borealis_exceptions.BorealisFileTypeError(
                self.filename, self.borealis_filetype)

    def read_bfiq(self) -> dict:
        """
        Reads Borealis bfiq file

        Returns
        -------
        records: OrderedDict{dict}
            records of beamformed iq data. Keys are first sequence timestamp
            (in ms since epoch).
        """
        pydarn_log.info("Reading Borealis bfiq file: {}"
                        "".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisBfiq.site_single_element_types()
        dataset_types = borealis_formats.BorealisBfiq.site_array_dtypes()
        self._read_borealis_records(attribute_types, dataset_types)
        return self._records

    def read_rawacf(self) -> dict:
        """
        Reads Borealis rawacf file

        Returns
        -------
        records: OrderedDict{dict}
            records of borealis rawacf data. Keys are first sequence timestamp
            (in ms since epoch).
        """
        pydarn_log.info(
            "Reading Borealis rawacf file: {}".format(self.filename))

        attribute_types = \
            borealis_formats.BorealisRawacf.site_single_element_types()
        dataset_types = borealis_formats.BorealisRawacf.site_array_dtypes()
        self._read_borealis_records(attribute_types, dataset_types)
        return self._records

    def read_antennas_iq(self) -> dict:
        """
        Reads Borealis antennas_iq file

        Returns
        -------
        records: OrderedDict{dict}
            records of borealis antennas iq data. Keys are first sequence
            timestamp (in ms since epoch).
        """
        pydarn_log.info("Reading Borealis antennas_iq file: {}"
                        "".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisAntennasIq.site_single_element_types()
        dataset_types = borealis_formats.BorealisAntennasIq.site_array_dtypes()
        self._read_borealis_records(attribute_types, dataset_types)
        return self._records

    def read_rawrf(self) -> dict:
        """
        Reads Borealis rawrf file

        Returns
        -------
        records: OrderedDict{dict}
            records of borealis rawrf data. Keys are first sequence timestamp
            (in ms since epoch).
        """
        pydarn_log.info("Reading Borealis rawrf file: {}"
                        "".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisRawrf.site_single_element_types()
        dataset_types = borealis_formats.BorealisRawrf.site_array_dtypes()
        self._read_borealis_records(attribute_types, dataset_types)
        return self._records

    def _read_borealis_records(self, attribute_types: dict,
                               dataset_types: dict):
        """
        Read the entire file while checking all data fields.

        Several Borealis field checks are done to insure the integrity of the
        file.

        Parameters
        ----------
        attribute_types: dict
            Dictionary with the required types for the attributes in the file.
        dataset_types: dict
            Dictionary with the require dtypes for the numpy arrays in the
            file.

        Raises
        ------
        OSError: file does not exist
        BorealisFieldMissingError - when a field is missing from the Borealis
                                file/stream type
        BorealisExtraFieldError - when an extra field is present in the
                                Borealis file/stream type
        BorealisDataFormatTypeError - when a field has the incorrect
                                field type for the Borealis file/stream type

        See Also
        --------
        BorealisUtilities
        """
        records = dd.io.load(self.filename)
        BorealisUtilities.check_records(self.filename, records,
                                        attribute_types, dataset_types)

        self._records = OrderedDict(sorted(records.items()))


class BorealisSiteWrite():
    """
    Class for writing Borealis 'site' filetypes. Site files are stored in
    a record-by-record style.

    See Also
    --------
    BorealisRawacf
    BorealisBfiq
    BorealisAntennasIq
    BorealisRawrf

    Attributes
    ----------
    filename: str
        The filename of the Borealis HDF5 file being read.
    borealis_filetype
        Borealis filetype. Currently supported:
        - bfiq
        - antennas_iq
        - rawacf
        - rawrf
    record_names: list(str)
    records: OrderedDict{dict}
    arrays: dict
    compression: str
        The type of compression to write the file as. Default None.
        Default is no compression to match site-structure files
        created on site by the Borealis radar. They are not compressed
        as they are appended to record by record.
    """
    def __init__(self, filename: str,
                 borealis_records: OrderedDict,
                 borealis_filetype: str,
                 hdf5_compression: Union[str, None] = None):
        """
        Write borealis records to a file.

        Parameters
        ----------
        filename: str
            Name of the file the user wants to write to
        borealis_records: OrderedDict{dict}
            OrderedDict of borealis records, where keys are the
            record name
        borealis_filetype
            filetype to write as. Currently supported:
                - bfiq
                - rawacf
                - antennas_iq
                - rawrf
        hdf5_compression
            String representing hdf5 compression type. Default None.
        """
        self._records = borealis_records
        self.borealis_filetype = borealis_filetype
        self.filename = filename
        self.compression = hdf5_compression
        self._record_names = sorted(list(borealis_records.keys()))
        # list of group keys for partial write
        self.write_file()

    def __repr__(self):
        """For representation of the class object"""

        return "{class_name}({filename})"\
               "".format(class_name=self.__class__.__name__,
                         filename=self.filename)

    def __str__(self):
        """For printing of the class object"""

        return "Writing to filename: {filename} at record name: "\
               "".format(filename=self.filename)

    @property
    def record_names(self):
        """
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms since
        epoch), and are equal to the group names in the site file types.
        """
        return self._record_names

    @property
    def records(self):
        """
        The Borealis data in a dictionary of records, according to the
        site file format.
        """
        return self._records

    @property
    def arrays(self):
        """
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
        """
        return BorealisRestructureUtilities.borealis_site_to_array_dict(
                                            self.filename, self.records,
                                            self.borealis_filetype)

    def write_file(self) -> str:
        """
        Write Borealis records to a file given filetype.

        Raises
        ------
        BorealisFileTypeError
        """

        if self.borealis_filetype == 'bfiq':
            self.write_bfiq()
        elif self.borealis_filetype == 'rawacf':
            self.write_rawacf()
        elif self.borealis_filetype == 'antennas_iq':
            self.write_antennas_iq()
        elif self.borealis_filetype == 'rawrf':
            self.write_rawrf()
        else:
            raise borealis_exceptions.\
                    BorealisFileTypeError(self.filename,
                                          self.borealis_filetype)

    def write_bfiq(self) -> str:
        """
        Writes Borealis bfiq file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info(
            "Writing Borealis bfiq file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisBfiq.site_single_element_types()
        dataset_types = borealis_formats.BorealisBfiq.site_array_dtypes()
        self._write_borealis_records(attribute_types, dataset_types)
        return self.filename

    def write_rawacf(self) -> str:
        """
        Writes Borealis rawacf file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info(
            "Writing Borealis rawacf file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisRawacf.site_single_element_types()
        dataset_types = borealis_formats.BorealisRawacf.site_array_dtypes()
        self._write_borealis_records(attribute_types, dataset_types)
        return self.filename

    def write_antennas_iq(self) -> str:
        """
        Writes Borealis antennas_iq file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info(
            "Writing Borealis antennas_iq file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisAntennasIq.site_single_element_types()
        dataset_types = borealis_formats.BorealisAntennasIq.site_array_dtypes()
        self._write_borealis_records(attribute_types, dataset_types)
        return self.filename

    def write_rawrf(self) -> str:
        """
        Writes Borealis rawrf file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info(
            "Writing Borealis rawrf file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisRawrf.site_single_element_types()
        dataset_types = borealis_formats.BorealisRawrf.site_array_dtypes()
        self._write_borealis_records(attribute_types, dataset_types)
        return self.filename

    def _write_borealis_records(self, attribute_types: dict,
                                dataset_types: dict):
        """
        Write the file in site style after checking records.

        Several Borealis field checks are done to insure the integrity of the
        file.

        Parameters
        ----------
        attributes_type_dict: dict
            Dictionary with the required types for the attributes in the file.
        datasets_type_dict: dict
            Dictionary with the require dtypes for the numpy arrays in the
            file.

        Raises
        ------
        BorealisFieldMissingError - when a field is missing from the Borealis
                                file/stream type
        BorealisExtraFieldError - when an extra field is present in the
                                Borealis file/stream type
        BorealisDataFormatTypeError - when a field has the incorrect
                                field type for the Borealis file/stream type

        See Also
        --------
        BorealisUtilities
        """
        Path(self.filename).touch()
        BorealisUtilities.check_records(self.filename, self.records,
                                        attribute_types, dataset_types)

        # use external h5copy utility to move new record into 2hr file.
        cp_cmd = 'h5copy -i {newfile} -o {full_file} -s {dtstr} -d {dtstr}'
        warnings.filterwarnings("ignore")
        # Must use temporary file to append to a file; writing entire
        # dictionary at once also doesn't work so this is required.
        tmp_filename = self.filename + '.tmp'
        Path(tmp_filename).touch()
        for group_name, group_dict in self.records.items():
            dd.io.save(tmp_filename, {str(group_name): group_dict},
                       compression=self.compression)
            cmd = cp_cmd.format(newfile=tmp_filename, full_file=self.filename,
                                dtstr='/'+str(group_name))

            sp.call(cmd.split())
            os.remove(tmp_filename)



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author(s): Marina Schmidt
import logging

pydarn_log = logging.getLogger('pydarn')


class HardwareFileNotFoundError(Exception):
    """
    This error is raised when a hardware file for the provided
    abbreviation is not found.
    """
    def __init__(self, abbrev):
        self.abbreviation = abbrev
        self.message = "Hardware file for {} radar was not found."\
            " Please insure the abbreviation is correct and there"\
            " exists a hardware file for it by checking:"\
            " https://github.com/vtsuperdarn/hdw.dat"\
            "".format(self.abbreviation)

        super().__init__(self.message)
        pydarn_log.error(self.message)



# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marci Detwiller
"""
This file contains classes for reading, and
writing of Borealis array file types. Array file
types are restructured Borealis files. Files are restructured
after being written by the radar to condense all information
from that time period into arrays. The other type of
Borealis file is the site file. Site file types
are Borealis site files, ie. stored in a record-by-record
fashion, before being converted to array types.

Classes
-------
BorealisArrayRead: Reads array-style Borealis SuperDARN file types (hdf5).
    These are the files commonly distributed and available.
BorealisArrayWrite: Writes array-style Borealis SuperDARN file types (hdf5).
    These are the files commonly distributed and available.

Exceptions
----------
BorealisFileTypeError
BorealisFieldMissingError
BorealisExtraFieldError
BorealisDataFormatTypeError
BorealisNumberOfRecordsError

See Also
--------
BorealisUtilities
BorealisSiteRead
BorealisSiteWrite
BorealisRestructureUtilities

For more information on Borealis data files and how they convert to SDarn
files, see: https://borealis.readthedocs.io/en/latest/
"""
import deepdish as dd
import logging

from typing import List

from pydarn import borealis_exceptions, borealis_formats

from .borealis_utilities import BorealisUtilities
from .restructure_borealis import BorealisRestructureUtilities

pydarn_log = logging.getLogger('pydarn')


class BorealisArrayRead():
    """
    Class for reading Borealis 'array' filetypes. Array files have
    been restructured to no longer be read in record-by-record
    style.

    See Also
    --------
    BorealisRawacf
    BorealisBfiq
    BorealisAntennasIq
    BorealisRawrf

    Attributes
    ----------
    filename: str
        The filename of the Borealis HDF5 file being read.
    borealis_filetype: str
        The type of Borealis file. Restructured types include:
        'bfiq'
        'antennas_iq'
        'rawacf'
    record_names: list(str)
    records: dict
    arrays: dict
    """

    def __init__(self, filename: str, borealis_filetype: str):
        """
        Reads Borealis array file types into a dictionary.

        Parameters
        ----------
        filename: str
            file name containing Borealis hdf5 data.
        borealis_filetype: str
            The type of Borealis file. Types include:
            'bfiq'
            'antennas_iq'
            'rawacf'
            'rawrf'

        Raises
        ------
        OSError
            Unable to open file
        """
        self.filename = filename

        if borealis_filetype not in ['bfiq', 'antennas_iq', 'rawacf', 'rawrf']:
            raise borealis_exceptions.BorealisFileTypeError(
                self.filename, borealis_filetype)
        self.borealis_filetype = borealis_filetype

        # Records are private to avoid tampering.
        self._arrays = {}
        self.read_file()

    def __repr__(self):
        """ for representation of the class object"""
        # __class__.__name__ allows to grab the class name such that
        # when a class inherits this one, the class name will be the child
        # class and not the parent class
        return "{class_name}({filename}, {borealis_filetype})"\
            "".format(class_name=self.__class__.__name__,
                      filename=self.filename,
                      borealis_filetype=self.borealis_filetype)

    def __str__(self):
        """ for printing of the class object"""

        return "Reading from {filename}"\
            "".format(filename=self.filename)

    @property
    def record_names(self):
        """
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms), and
        are equal to the group names in the site file types.
        """
        return sorted(list(self.records.keys()))

    @property
    def records(self):
        """
        The Borealis data in a dictionary of records, according to the
        site file format.
        """
        return BorealisRestructureUtilities.borealis_array_to_site_dict(
                                            self.filename, self.arrays,
                                            self.borealis_filetype)

    @property
    def arrays(self):
        """
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
        """
        return self._arrays

    def read_file(self) -> dict:
        """
        Reads the specified Borealis file using the other functions for
        the proper file type.

        Reads the entire file.

        See Also
        --------
        read_bfiq
        read_rawacf
        read_antennas_iq

        Returns
        -------
        arrays: dict
            borealis data dictionary. Keys are data field names and
            unshared fields have a first dimension = number of records
            in the file.

        Raises
        ------
        BorealisFileTypeError
        """
        if self.borealis_filetype == 'bfiq':
            return self.read_bfiq()
        elif self.borealis_filetype == 'rawacf':
            return self.read_rawacf()
        elif self.borealis_filetype == 'antennas_iq':
            return self.read_antennas_iq()
        else:
            raise borealis_exceptions.BorealisFileTypeError(
                self.filename, self.borealis_filetype)

    def read_bfiq(self) -> dict:
        """
        Reads Borealis bfiq file that has been structured into arrays.

        Returns
        -------
        arrays: dict
            The Borealis data in a dictionary of arrays, according to the
            restructured array file format.
        """
        pydarn_log.info("Reading Borealis bfiq file: {}"
                        "".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisBfiq.array_single_element_types()
        dataset_types = borealis_formats.BorealisBfiq.array_array_dtypes()
        unshared_fields = borealis_formats.BorealisBfiq.unshared_fields + \
            borealis_formats.BorealisBfiq.array_only_fields
        self._read_borealis_arrays(attribute_types,
                                   dataset_types, unshared_fields)
        return self._arrays

    def read_rawacf(self) -> dict:
        """
        Reads Borealis rawacf file that has been structured into arrays.

        Returns
        -------
        arrays: dict
            The Borealis data in a dictionary of arrays, according to the
            restructured array file format.
        """
        pydarn_log.info(
            "Reading Borealis rawacf file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisRawacf.array_single_element_types()
        dataset_types = borealis_formats.BorealisRawacf.array_array_dtypes()
        unshared_fields = borealis_formats.BorealisRawacf.unshared_fields + \
            borealis_formats.BorealisRawacf.array_only_fields
        self._read_borealis_arrays(attribute_types, dataset_types,
                                   unshared_fields)
        return self._arrays

    def read_antennas_iq(self) -> dict:
        """
        Reads Borealis antennas_iq file that has been structured into arrays.

        Returns
        -------
        arrays: dict
            The Borealis data in a dictionary of arrays, according to the
            restructured array file format.
        """
        pydarn_log.info("Reading Borealis antennas_iq file: {}"
                        "".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisAntennasIq.array_single_element_types()
        dataset_types = \
            borealis_formats.BorealisAntennasIq.array_array_dtypes()
        unshared_fields = \
            borealis_formats.BorealisAntennasIq.unshared_fields + \
            borealis_formats.BorealisAntennasIq.array_only_fields
        self._read_borealis_arrays(attribute_types, dataset_types,
                                   unshared_fields)
        return self._arrays

    def _read_borealis_arrays(self, attribute_types: dict,
                              dataset_types: dict,
                              unshared_fields: List[str]):
        """
        Read the entire file while checking all data fields.

        Parameters
        ----------
        attribute_types: dict
            Dictionary with the required types for the attributes in the file.
        dataset_types: dict
            Dictionary with the require dtypes for the numpy arrays in the
            file.
        unshared_fields: List[str]
            List of fields that are not shared between the records and
            therefore should be an array with first dimension = number of
            records

        Raises
        ------
        BorealisFieldMissingError - when a field is missing from the Borealis
                                file
        BorealisExtraFieldError - when an extra field is present in the
                                Borealis file
        BorealisDataFormatTypeError - when a field has the incorrect
                                field type for the Borealis file
        BorealisNumberOfRecordsError - when the number of records cannot
                                be discerned from the arrays

        See Also
        --------
        BorealisUtilities
        """
        arrays = dd.io.load(self.filename)
        BorealisUtilities.check_arrays(self.filename, arrays,
                                       attribute_types, dataset_types,
                                       unshared_fields)
        self._arrays = arrays


class BorealisArrayWrite():
    """
    Class for writing Borealis 'array' filetypes. Array files have
    been restructured to no longer be read in record-by-record
    style.

    See Also
    --------
    BorealisRawacf
    BorealisBfiq
    BorealisAntennasIq
    BorealisRawrf

    Attributes
    ----------
    filename: str
        The filename of the Borealis HDF5 file being written.
    borealis_filetype: str
        The type of Borealis file. Restructured types include:
        'bfiq'
        'antennas_iq'
        'rawacf'
    record_names: list(str)
    records: dict
    arrays: dict
    compression: str
        The type of compression to write the file as. Default zlib.
        zlib is hdf5 default compression for fast reading. We want
        the fastest read possible for downstream users with this
        structure style.
    """

    def __init__(self, filename: str, borealis_arrays: dict,
                 borealis_filetype: str, hdf5_compression: str = 'zlib'):
        """
        Write borealis arrays to an array restructured file.

        Parameters
        ----------
        filename: str
            Name of the file the user wants to write to
        borealis_arrays: dict
            borealis data dictionary. Keys are data field names and
            unshared fields have a first dimension = number of records
            in the file.
        borealis_filetype: str
            The type of Borealis file. Restructured types include:
            'bfiq'
            'antennas_iq'
            'rawacf'
        hdf5_compression: str
            String representing hdf5 compression type. Default zlib.
        """
        self.filename = filename
        self._arrays = borealis_arrays
        self.borealis_filetype = borealis_filetype
        self.compression = hdf5_compression
        self.write_file()

    def __repr__(self):
        """For representation of the class object"""

        return "{class_name}({filename}, {current_record_name})"\
               "".format(class_name=self.__class__.__name__,
                         filename=self.filename,
                         current_record_name=self.current_record_name)

    def __str__(self):
        """For printing of the class object"""

        return "Writing to filename: {filename} at record name: "\
               "{current_record_name}"\
               "".format(filename=self.filename,
                         current_record_name=self.current_record_name)

    @property
    def record_names(self):
        """
        A sorted list of the set of record names in the HDF5 file read.
        These correspond to Borealis file record write times (in ms), and
        are equal to the group names in the site file types.
        """
        return sorted(list(self.records.keys()))

    @property
    def records(self):
        """
        The Borealis data in a dictionary of records, according to the
        site file format.
        """
        return BorealisRestructureUtilities.\
            borealis_array_to_site_dict(self.filename, self.arrays,
                                        self.borealis_filetype)

    @property
    def arrays(self):
        """
        The Borealis data in a dictionary of arrays, according to the
        restructured array file format.
        """
        return self._arrays

    def write_file(self) -> str:
        """
        Write Borealis records to a file given filetype.

        Raises
        ------
        BorealisFileTypeError
        """

        if self.borealis_filetype == 'bfiq':
            self.write_bfiq()
        elif self.borealis_filetype == 'rawacf':
            self.write_rawacf()
        elif self.borealis_filetype == 'antennas_iq':
            self.write_antennas_iq()
        else:
            raise borealis_exceptions.\
                BorealisFileTypeError(self.filename,
                                      self.borealis_filetype)

    def write_bfiq(self) -> str:
        """
        Writes Borealis bfiq file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info("Writing Borealis bfiq file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisBfiq.array_single_element_types()
        dataset_types = borealis_formats.BorealisBfiq.array_array_dtypes()
        unshared_fields = borealis_formats.BorealisBfiq.unshared_fields + \
            borealis_formats.BorealisBfiq.array_only_fields
        self._write_borealis_arrays(attribute_types, dataset_types,
                                    unshared_fields)
        return self.filename

    def write_rawacf(self) -> str:
        """
        Writes Borealis rawacf file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info("Writing Borealis"
                        " rawacf file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisRawacf.array_single_element_types()
        dataset_types = borealis_formats.BorealisRawacf.array_array_dtypes()
        unshared_fields = borealis_formats.BorealisRawacf.unshared_fields + \
            borealis_formats.BorealisRawacf.array_only_fields
        self._write_borealis_arrays(attribute_types, dataset_types,
                                    unshared_fields)
        return self.filename

    def write_antennas_iq(self) -> str:
        """
        Writes Borealis antennas_iq file

        Returns
        -------
        filename: str
            Filename of written file.
        """
        pydarn_log.info("Writing Borealis"
                        " antennas_iq file: {}".format(self.filename))
        attribute_types = \
            borealis_formats.BorealisAntennasIq.array_single_element_types()
        dataset_types = \
            borealis_formats.BorealisAntennasIq.array_array_dtypes()
        unshared_fields = \
            borealis_formats.BorealisAntennasIq.unshared_fields + \
            borealis_formats.BorealisAntennasIq.array_only_fields
        self._write_borealis_arrays(attribute_types, dataset_types,
                                    unshared_fields)
        return self.filename

    def _write_borealis_arrays(self, attribute_types: dict,
                               dataset_types: dict,
                               unshared_fields: List[str]):
        """
        Write the entire file while checking all data fields.

        Parameters
        ----------
        attribute_types: dict
            Dictionary with the required types for the attributes in the file.
        dataset_types: dict
            Dictionary with the require dtypes for the numpy arrays in the
            file.
        unshared_fields: List[str]
            List of fields that are not shared between the records and
            therefore should be an array with first dimension = number of
            records

        Raises
        ------
        BorealisFieldMissingError - when a field is missing from the Borealis
                                file
        BorealisExtraFieldError - when an extra field is present in the
                                Borealis file
        BorealisDataFormatTypeError - when a field has the incorrect
                                field type for the Borealis file
        BorealisNumberOfRecordsError - when the number of records cannot
                                be discerned from the arrays

        See Also
        --------
        BorealisUtilities:
        """
        BorealisUtilities.check_arrays(self.filename, self.arrays,
                                       attribute_types, dataset_types,
                                       unshared_fields)
        dd.io.save(self.filename, self.arrays, compression=self.compression)
